import textwrap
from typing import Union, TYPE_CHECKING, Type, List

from antlr4.Token import CommonToken
from antlr4.tree.Tree import TerminalNodeImpl

if TYPE_CHECKING:
    from typing import TypeVar
    T = TypeVar('T')

def is_pow2(x: int) -> bool:
    return (x > 0) and ((x & (x - 1)) == 0)


def roundup_pow2(x: int) -> int:
    return 1<<(x-1).bit_length()

def roundup_to(x: int, n: int) -> int:
    if x % n:
        return (x//n + 1) * n
    else:
        return (x//n) * n

def get_ID_text(token: Union[CommonToken, TerminalNodeImpl]) -> str:
    """
    Get the text from the ID token.
    Strips off leading slash escape if present
    """
    if isinstance(token, CommonToken):
        text = token.text
    else:
        text = token.getText()

    text = text.lstrip('\\')
    return text

def truncate_int(v: int, width: int) -> int:
    mask = (1 << width) - 1
    return v & mask

def dedent_text(s: str) -> str:
    """
    Remove any common indentation, ignoring indentation state of the first
    line of text.
    """
    s = s.strip()
    linelist = s.splitlines()
    if len(linelist) >= 2:
        s = (
            linelist[0]
            + "\n"
            + textwrap.dedent("\n".join(linelist[1:]))
        )
    return s

def get_all_subclasses(cls: Type['T']) -> List[Type['T']]:
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__()
        for g in get_all_subclasses(s)
    ]



from typing import TYPE_CHECKING, Optional, Any, Dict
from copy import deepcopy

from .value_normalization import normalize

if TYPE_CHECKING:
    from .. import rdltypes
    from ..ast import ASTNode

class Parameter:
    def __init__(self, param_type: 'rdltypes.PreElabRDLType', name: str, default_expr: Optional['ASTNode']=None):
        self.name = name
        self.param_type = param_type

        self.expr = default_expr

        # Stores the evaluated result of self.expr so that subsequent queries do
        # not need to repeatedly re-evaluate it
        self._value = None # type: Any


    def __deepcopy__(self, memo: Dict[int, Any]) -> 'Parameter':
        """
        Never implicitly deepcopy a Parameter object. Parameters are pointed to
        by reference so that value overrides propagate through the expression
        tree correctly.
        """
        return self


    def _copy_for_inst(self, memo: Dict[int, Any]) -> 'Parameter':
        """
        Explicitly deepcopy this object. This is only used when instantiating
        the enclosing component
        """
        if id(self) in memo:
            return memo[id(self)]

        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))

        # Reset cached value when copying
        result._value = None

        return result


    def get_value(self) -> Any:
        """
        Evaluate self.expr to get the parameter's value
        """
        if (self._value is None) and (self.expr is not None):
            self._value = self.expr.get_value()

        return self._value


    def get_normalized_parameter(self) -> str:
        """
        Converts the parameter to the normalized type name segment as defined in
        SystemRDL 2.0 Section 5.1.1.4-c

        Returns the whole parameter string:
            <parameter name> + "_" + <normalized value>
        """
        return self.name + "_" + normalize(self.get_value())



import operator
import functools
from copy import deepcopy
from collections import OrderedDict
from typing import Optional, List, Dict, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from typing import TypeVar
    from .core.parameter import Parameter
    from .source_ref import SourceRefBase

    ComponentClass = TypeVar('ComponentClass', bound='Component')

class Component:
    """
    Base class for all component types

    .. inheritance-diagram:: systemrdl.component
        :top-classes: ~Component
    """

    def __init__(self) -> None:
        #------------------------------
        # Component definition
        #------------------------------

        #: Reference to parent :class:`~systemrdl.component.Component`
        #: definition that lexically encloses this component type definition.
        #:
        #: May remain ``None`` if scope is not known or not applicable.
        #:
        #: .. note::
        #:      This represents the parent *lexical* scope! This does *not*
        #:      refer to the hierarchical parent of this component.
        self.parent_scope = None # type: Component

        # Name of this component's declaration scope
        # This field is only valid in non-instantiated components (referenced
        # via an instance's original_def)
        # If declaration was anonymous, inherits the first instance's name,
        # otherwise it contains the original type name.
        self._scope_name = None # type: str

        #: Named definition identifier.
        #: If declaration was anonymous, instantiation type names inherit
        #: the first instance's name.
        #: The type name of parameterized components is normalized based on the
        #: instance's parameter values.
        #:
        #: If this is a non-instantiated component (referenced via an
        #: instance's original_def), then it will contain either the original
        #: type name, or None if the declaration was anonymous.
        #:
        #: Importers may leave this as ``None`` if an appropriate type name
        #: cannot be imported.
        self.type_name = None # type: Optional[str]

        #: List of :class:`~systemrdl.component.Component` instances that are
        #: direct descendants of this component.
        #:
        #: Child components are sorted as follows:
        #:
        #: - Signals first
        #: - All other components follow.
        #: - AddressableComponents are sorted by ascending base_addr
        #: - Fields are sorted by ascending low bit
        self.children = [] # type: List[Component]

        # Parameters of this component definition.
        # These are listed in the order that they were defined
        self.parameters = [] # type: List[Parameter]

        # Properties applied to this component
        self.properties = {} # type: Dict[str, Any]

        #: :ref:`api_src_ref` for each explicit property assignment (if available)
        self.property_src_ref = {} # type: Dict[str, SourceRefBase]

        #: :ref:`api_src_ref` for the component definition
        self.def_src_ref = None # type: Optional[SourceRefBase]

        #------------------------------
        # Component instantiation
        #------------------------------
        #: If instantiated, set to True
        self.is_instance = False # type: bool

        #: Name of instantiated element
        self.inst_name = None # type: Optional[str]

        #: Reference to original :class:`~systemrdl.component.Component`
        #: definition this instance is derived from.
        #:
        #: Importers may leave this as ``None`` if appropriate.
        self.original_def = None # type: Optional[Component]

        #: True if instance type is external. False if internal.
        self.external = None # type: bool

        #: :ref:`api_src_ref` for the component instantiation.
        self.inst_src_ref = None # type: Optional[SourceRefBase]

        #------------------------------
        # List of property names that were assigned via a dynamic property
        # assignment.
        self._dyn_assigned_props = [] # type: List[str]

        # List of child instances that were assigned "through" this component,
        # from outside this component's scope.
        self._dyn_assigned_children = [] # type: List[str]


    def _copy_for_inst(self: 'ComponentClass', memo: Dict[int, Any]) -> 'ComponentClass':
        """
        Make a copy of the component tree in order to instantiate it.

        This is subtly different from a normal deepcopy since it ensures that
        references within the component tree are deepcopied, while references
        to external parameters are copied by reference.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        # First, explicitly copy all parameter objects
        result.parameters = [param._copy_for_inst(memo) for param in self.parameters]

        # Ensure child components get copied first
        result.children = [child._copy_for_inst(memo) for child in self.children]

        # Finally, continue deepcopying everything else
        copy_by_ref = {"original_def", "parent_scope", "comp_defs"}
        skip = {"parameters", "children"}
        for k, v in self.__dict__.items():
            if k in skip:
                continue
            if k in copy_by_ref:
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def __deepcopy__(self: 'ComponentClass', memo: Dict[int, Any]) -> 'ComponentClass':
        return self._copy_for_inst(memo)

    def __repr__(self) -> str:
        if self.is_instance:
            name_str = f"{self.inst_name} ({self.type_name})"
        else:
            name_str = self.type_name

        return f"<{self.__class__.__qualname__} {name_str} at {id(self):#x}>"


    def get_child_by_name(self, inst_name: str) -> Optional['Component']:
        for child in self.children:
            if child.inst_name == inst_name:
                return child
        return None


    def get_scope_path(self, scope_separator: str="::") -> Optional[str]:
        """
        Generate a string that represents this component's declaration namespace
        scope.

        Returns ``None`` if scope is not known or not applicable.

        For example, the following SystemRDL snippet:

        .. code-block:: systemrdl

            reg my_reg_t {
                field {} x;
            };

            addrmap top {
                my_reg_t foo;
                reg my_other_reg_t {
                    field {} y;
                } bar;
                reg {
                    field {} z;
                } baz, xyz;
            };

        ... results in:

        * Field ``x`` of hierarchical path ``top.foo.x`` was declared in the
          lexical scope ``my_reg_t``
        * Field ``y`` of hierarchical path ``top.bar.y`` was declared in the
          lexical scope ``top::my_other_reg_t``
        * Both fields ``z`` of hierarchical paths ``top.baz.z`` and ``top.xyz.z``
          were declared in the same lexical scope ``top::baz``
        * Register ``foo`` was declared in the root scope which is represented
          by an empty string.

        Parameters
        ----------
        scope_separator: str
            Override the separator between namespace scopes
        """
        if self.parent_scope is None:
            # Importer likely never set the scope
            return None
        elif isinstance(self.parent_scope, Root):
            # Declaration was in root scope
            return ""
        else:
            # Get parent definition's scope path
            parent_path = self.parent_scope.get_scope_path(scope_separator)

            # Extend it with its scope name
            if parent_path:
                return(
                    parent_path
                    + scope_separator
                    + self.parent_scope._scope_name
                )
            else:
                return self.parent_scope._scope_name


class AddressableComponent(Component):
    """
    Base class for all components that can have an address
    """

    def __init__(self) -> None:
        super().__init__()
        #------------------------------
        # Component instantiation
        #------------------------------
        #: Address offset from the parent component.
        #: If left as None, compiler will resolve with inferred value.
        self.addr_offset = None # type: int

        #: Address alignment if explicitly assigned by user.
        self.addr_align = None # type: Optional[int]

        #------------------------------
        # Array Properties
        #------------------------------
        #: If true, then ``array_dimensions`` and ``array_stride`` are valid.
        self.is_array = False # type: bool

        #: List of sizes for each array dimension.
        #: Last item in the list iterates the most frequently.
        self.array_dimensions = None # type: Optional[List[int]]

        #: Address offset between array elements.
        #: If left as None, compiler will resolve with inferred value.
        self.array_stride = None # type: Optional[int]


    @property
    def n_elements(self) -> int:
        """
        Total number of array elements.
        If array is multidimensional, array is flattened.
        Returns 1 if not an array.
        """
        if self.is_array:
            assert isinstance(self.array_dimensions, list)
            return functools.reduce(operator.mul, self.array_dimensions)
        else:
            return 1

class VectorComponent(Component):
    """
    Base class for all components that are vector-like
    """

    def __init__(self) -> None:
        super().__init__()
        #------------------------------
        # Component instantiation
        #------------------------------
        #: Width of vector in bits
        self.width = None # type: int

        #: Bit position of most significant bit
        self.msb = None # type: int
        #: Bit position of least significant bit
        self.lsb = None # type: int

        #: High index of bit range
        self.high = None # type: int
        #: Low index of bit range
        self.low = None # type: int

#===============================================================================
class Root(Component):
    """
    Meta-component used by compiler to represent the root scope
    """
    def __init__(self) -> None:
        super().__init__()
        #: Dictionary of :class:`~systemrdl.component.Component` definitions in
        #: the global root scope.
        self.comp_defs = OrderedDict() # type: Dict[str, Component]

class Signal(VectorComponent):
    pass

class Field(VectorComponent):
    pass

class Reg(AddressableComponent):
    def __init__(self) -> None:
        super().__init__()

        #: If true, fields are to be arranged in msb0 order
        #:
        #: .. versionadded:: 1.7
        self.is_msb0_order = False # type: bool

        # List of register inst names that are aliases of this primary
        # Due to limitations in RDL grammar, these can only be siblings in the hierarchy
        # so names are unambiguous
        self._alias_names = [] # type: List[str]

        #------------------------------
        # Alias Register
        #------------------------------
        #: If true, then ``alias_primary_inst`` is valid
        self.is_alias = False # type: bool

        #: Reference to primary register :class:`~systemrdl.component.Component`
        #: instance
        self.alias_primary_inst = None # type: Optional[Reg]

class Regfile(AddressableComponent):
    pass

class Addrmap(AddressableComponent):
    pass

class Mem(AddressableComponent):
    pass



from typing import Any, Set, Type, Tuple, TYPE_CHECKING, Optional, Dict

from .. import node as m_node
from ..ast.ast_node import ASTNode
from .. import rdltypes
from ..ast.cast import AssignmentCast, is_castable
from ..core.helpers import get_all_subclasses

if TYPE_CHECKING:
    from .. import component as comp
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

#===============================================================================
# Base property
#===============================================================================
class PropertyRule:
    # Set of components this property can be bound to
    bindable_to = set() # type: Set[Type[comp.Component]]

    # List of valid assignment types. In order of cast preference
    valid_types = tuple() # type: Tuple[Any, ...]

    # Default value if not assigned
    default = None # type: Any

    # Whether dynamic assignments are allowed to be made to this property
    dyn_assign_allowed= True # type: bool

    # Group string in which this property is mutually exclusive
    mutex_group = None # type: Optional[str]


    def __init__(self, env: 'RDLEnvironment'):
        self.env = env


    @classmethod
    def get_name_cls(cls) -> str:
        return cls.__name__.replace("Prop_", "")


    def get_name(self) -> str:
        return self.get_name_cls()


    def assign_value(self, comp_def: 'comp.Component', value: Any, src_ref: 'SourceRefBase') -> None:
        """
        Used by the compiler for either local or dynamic prop assignments
        This does the following:
            - Check that the property is allowed in this component
            - Check if the value being assigned is compatible
            - Assign the property, as well as any side-effects
                subclasses extend this to define prop-specific side-effects
        """

        # Check if property is allowed in this component
        if type(comp_def) not in self.bindable_to:
            self.env.msg.fatal(
                "The property '%s' is not valid for '%s' components"
                % (self.get_name(), type(comp_def).__name__.lower()),
                src_ref
            )

        # Property assignments with no rhs show up as None here
        # For built-in properties, this implies a True value
        if value is None:
            value = True

        # unpack true type of value
        # Contents of value can be:
        #   - An expression (instance of an ASTNode subclass)
        #   - Direct assignment of a type-compatible value
        if isinstance(value, ASTNode):
            # Predict expected type after value would get evaluated
            assign_type = value.predict_type()
        elif rdltypes.is_user_enum(value):
            # Value is a user enum type derived from UserEnum.
            # Mark it as such
            assign_type = rdltypes.UserEnum
        else:
            # Value is already evaluated
            assign_type = type(value)

        # First check if the value's type is already directly compatible
        for valid_type in self.valid_types:
            if assign_type == valid_type:
                break
        else:
            # otherwise, cast to the first compatible type
            for valid_type in self.valid_types:
                if is_castable(assign_type, valid_type):
                    if isinstance(value, ASTNode):
                        # Found a type-compatible match. (first match is best match)
                        # Wrap the expression with an explicit assignment cast
                        value = AssignmentCast(self.env, src_ref, value, valid_type)
                    break
            else:
                self.env.msg.fatal(
                    "Incompatible assignment to property '%s'" % self.get_name(),
                    src_ref
                )

        # If the property belongs to a mutex group, wipe out any of its
        # counterpart properties
        if self.mutex_group is not None:
            for prop_name in _get_mutex_properties(self.mutex_group):
                if prop_name in comp_def.properties:
                    del comp_def.properties[prop_name]

        # Store the property
        comp_def.properties[self.get_name()] = value

        if src_ref is not None:
            comp_def.property_src_ref[self.get_name()] = src_ref


    def get_default(self, node: m_node.Node) -> Any:
        # pylint: disable=unused-argument
        """
        Used when the user queries a property, and it was not explicitly set.
        Default values are not always directly known. Sometimes they depend on
        one or more other properties.
        The base behavior will simply return the static variable's value.
        Properties with more complex rules can override this to implement
        other default value derivations
        """
        return self.default


    def validate(self, node: m_node.Node, value: Any) -> None:
        """
        Used during the validate phase after elaboration.
        Performs checks against the property's value
        """


    def _validate_ref_width(self, node: m_node.VectorNode, value: Any) -> None:
        """
        Helper function to check that if value is a vector-like reference,
        that its width matches the node.
        """
        if isinstance(value, m_node.VectorNode):
            if node.width != value.width:
                self.env.msg.error(
                    "%s '%s' references %s '%s''s value but they are not the same width (%d != %d)"
                    % (
                        type(node.inst).__name__.lower(), node.inst_name,
                        type(value.inst).__name__.lower(), value.inst_name,
                        node.width, value.width
                    ),
                    self.get_src_ref(node)
                )
        elif isinstance(value, rdltypes.PropertyReference) and value.width is not None:
            if node.width != value.width:
                self.env.msg.error(
                    "%s '%s' references property '%s->%s' but they are not the same width (%d != %d)"
                    % (
                        type(node.inst).__name__.lower(), node.inst_name,
                        value.node.inst_name, value.name,
                        node.width, value.width
                    ),
                    self.get_src_ref(node)
                )


    def _validate_ref_width_is_1(self, node: m_node.Node, value: Any) -> None:
        """
        Helper function to check that if value is a vector-like reference,
        that its width is exactly 1.
        """
        if isinstance(value, m_node.VectorNode):
            if value.width != 1:
                self.env.msg.error(
                    "%s '%s' references %s '%s' but its width is not 1"
                    % (
                        type(node.inst).__name__.lower(), node.inst_name,
                        type(value.inst).__name__.lower(), value.inst_name
                    ),
                    self.get_src_ref(node)
                )
        elif isinstance(value, rdltypes.PropertyReference) and value.width is not None:
            if value.width != 1:
                self.env.msg.error(
                    "%s '%s' references property '%s->%s' but its width is not 1"
                    % (
                        type(node.inst).__name__.lower(), node.inst_name,
                        value.node.inst_name, value.name,
                    ),
                    self.get_src_ref(node)
                )


    def _validate_ref_is_present(self, node: m_node.Node, value: Any) -> None:
        """
        5.3.1-i: If a present instance includes references (e.g., signals), the
        referred objects need to also be present.
        """
        if isinstance(value, m_node.Node):
            if not value.get_property('ispresent'):
                self.env.msg.error(
                    "%s '%s' references %s '%s' but it is not present (ispresent=false)"
                    % (
                        type(node.inst).__name__.lower(), node.inst_name,
                        type(value.inst).__name__.lower(), value.inst_name
                    ),
                    self.get_src_ref(node)
                )
        elif isinstance(value, rdltypes.PropertyReference):
            if not value.node.get_property('ispresent'):
                self.env.msg.error(
                    "%s '%s' references '%s->%s' but it is not present (ispresent=false)"
                    % (
                        type(node.inst).__name__.lower(), node.inst_name,
                        value.node.inst_name, value.name
                    ),
                    self.get_src_ref(node)
                )

    def _validate_width_eq_or_smaller(self, node: m_node.VectorNode, value: Any) -> None:
        """
        Helper function to check that if value is a vector-like reference,
        that its width is equal or smaller than the node's width

        Used for 9.8.1: The increment/decrement value shall be equal to or
        smaller than the field's width.
        """
        if isinstance(value, int):
            if node.width < value.bit_length():
                self.env.msg.error(
                    "A counter's %s cannot reference a value wider than the counter itself."
                    % (self.get_name()),
                    self.get_src_ref(node)
                )
        elif isinstance(value, m_node.VectorNode):
            if node.width < value.width:
                self.env.msg.error(
                    "A counter's %s cannot reference a value wider than the counter itself."
                    % (self.get_name()),
                    self.get_src_ref(node)
                )
        elif isinstance(value, rdltypes.PropertyReference) and value.width is not None:
            if node.width < value.width:
                self.env.msg.error(
                    "A counter's %s cannot reference a value wider than the counter itself."
                    % (self.get_name()),
                    self.get_src_ref(node)
                )

    def get_src_ref(self, node: m_node.Node) -> 'SourceRefBase':
        return node.inst.property_src_ref.get(self.get_name(), node.inst.inst_src_ref)


#===============================================================================
class PropertyRuleBoolPair(PropertyRule):
    # Property name of the equivalent opposite
    opposite_property = ""

    def get_default(self, node: m_node.Node) -> bool:
        """
        If not explicitly set, check if the opposite was set first before returning
        default
        """
        # Check for explicit assignment to avoid infinite loop via get_default()
        if self.opposite_property in node.inst.properties:
            return not node.inst.properties[self.opposite_property]
        else:
            return self.default

#===============================================================================
_MUTEX_PROP_GROUPS = {} # type: Dict[str, Set[str]]
def _get_mutex_properties(group_name: str) -> Set[str]:
    # initialize cache for the first time if needed
    if not _MUTEX_PROP_GROUPS:
        for prop_cls in get_all_subclasses(PropertyRule):
            if prop_cls.__name__.startswith("Prop_"):
                if prop_cls.mutex_group is not None:
                    if prop_cls.mutex_group not in _MUTEX_PROP_GROUPS:
                        _MUTEX_PROP_GROUPS[prop_cls.mutex_group] = set()
                    _MUTEX_PROP_GROUPS[prop_cls.mutex_group].add(prop_cls.get_name_cls())

    return _MUTEX_PROP_GROUPS[group_name]



import re
from collections import OrderedDict

from ..parser.SystemRDLParser import SystemRDLParser
from .. import rdltypes
from .. import component as comp
from ..source_ref import src_ref_from_antlr

from .BaseVisitor import BaseVisitor
from .. import ast
from .parameter import Parameter
from .helpers import get_ID_text

class ExprVisitor(BaseVisitor):

    #---------------------------------------------------------------------------
    # Numerical Expressions
    #---------------------------------------------------------------------------
    _BinaryExpr_map = {
        SystemRDLParser.EXP     : ast.Exponent,
        SystemRDLParser.MULT    : ast.Mult,
        SystemRDLParser.DIV     : ast.Div,
        SystemRDLParser.MOD     : ast.Mod,
        SystemRDLParser.PLUS    : ast.Add,
        SystemRDLParser.MINUS   : ast.Sub,
        SystemRDLParser.LSHIFT  : ast.LShift,
        SystemRDLParser.RSHIFT  : ast.RShift,
        SystemRDLParser.LT      : ast.Lt,
        SystemRDLParser.LEQ     : ast.Leq,
        SystemRDLParser.GT      : ast.Gt,
        SystemRDLParser.GEQ     : ast.Geq,
        SystemRDLParser.EQ      : ast.Eq,
        SystemRDLParser.NEQ     : ast.Neq,
        SystemRDLParser.AND     : ast.BitwiseAnd,
        SystemRDLParser.XOR     : ast.BitwiseXor,
        SystemRDLParser.XNOR    : ast.BitwiseXnor,
        SystemRDLParser.OR      : ast.BitwiseOr,
        SystemRDLParser.BAND    : ast.BoolAnd,
        SystemRDLParser.BOR     : ast.BoolOr
    }

    _UnaryExpr_map = {
        SystemRDLParser.PLUS    : ast.UnaryPlus,
        SystemRDLParser.MINUS   : ast.UnaryMinus,
        SystemRDLParser.BNOT    : ast.BoolNot,
        SystemRDLParser.NOT     : ast.BitwiseInvert,
        SystemRDLParser.AND     : ast.AndReduce,
        SystemRDLParser.NAND    : ast.NandReduce,
        SystemRDLParser.OR      : ast.OrReduce,
        SystemRDLParser.NOR     : ast.NorReduce,
        SystemRDLParser.XOR     : ast.XorReduce,
        SystemRDLParser.XNOR    : ast.XnorReduce
    }

    # Visit a parse tree produced by SystemRDLParser#BinaryExpr.
    def visitBinaryExpr(self, ctx: SystemRDLParser.BinaryExprContext):
        l = self.visit(ctx.expr(0))
        r = self.visit(ctx.expr(1))
        expr_class = self._BinaryExpr_map[ctx.op.type]
        return expr_class(self.compiler.env, src_ref_from_antlr(ctx.op), l, r)


    # Visit a parse tree produced by SystemRDLParser#UnaryExpr.
    def visitUnaryExpr(self, ctx: SystemRDLParser.UnaryExprContext):
        n = self.visit(ctx.expr_primary())
        expr_class = self._UnaryExpr_map[ctx.op.type]
        return expr_class(self.compiler.env, src_ref_from_antlr(ctx.op), n)


    # Visit a parse tree produced by SystemRDLParser#TernaryExpr.
    def visitTernaryExpr(self, ctx: SystemRDLParser.TernaryExprContext):
        i = self.visit(ctx.expr(0))
        j = self.visit(ctx.expr(1))
        k = self.visit(ctx.expr(2))
        return ast.Conditional(self.compiler.env, src_ref_from_antlr(ctx.op), i, j, k)


    # Visit a parse tree produced by SystemRDLParser#paren_expr.
    def visitParen_expr(self, ctx: SystemRDLParser.Paren_exprContext):
        return self.visit(ctx.expr())

    #---------------------------------------------------------------------------
    # Numeric Literals
    #---------------------------------------------------------------------------
    # Visit a parse tree produced by SystemRDLParser#NumberInt.
    def visitNumberInt(self, ctx: SystemRDLParser.NumberIntContext):
        s = ctx.INT().getText()
        s = s.replace("_", "")
        return ast.IntLiteral(self.compiler.env, src_ref_from_antlr(ctx.INT()), int(s))


    # Visit a parse tree produced by SystemRDLParser#NumberHex.
    def visitNumberHex(self, ctx: SystemRDLParser.NumberHexContext):
        s = ctx.HEX_INT().getText()
        s = s.replace("_", "")
        return ast.IntLiteral(self.compiler.env, src_ref_from_antlr(ctx.HEX_INT()), int(s, 16))


    # Visit a parse tree produced by SystemRDLParser#NumberVerilog.
    def visitNumberVerilog(self, ctx: SystemRDLParser.NumberVerilogContext):
        s = ctx.VLOG_INT().getText()
        s = s.replace("_", "")
        m = re.fullmatch(r"(\d+)'(b|d|h)([\da-f]+)", s, re.IGNORECASE)
        width = int(m.group(1))
        basechar = m.group(2).lower()
        if basechar == "b":
            base = 2
        elif basechar == "d":
            base = 10
        else:
            base = 16

        val = int(m.group(3), base)

        if width < 1:
            self.msg.fatal(
                "Integer literal width must be greater than zero",
                src_ref_from_antlr(ctx.VLOG_INT())
            )

        if val.bit_length() > width:
            self.msg.fatal(
                "Value of integer literal exceeds the specified width",
                src_ref_from_antlr(ctx.VLOG_INT())
            )

        return ast.IntLiteral(self.compiler.env, src_ref_from_antlr(ctx.VLOG_INT()), val, width)


    # Visit a parse tree produced by SystemRDLParser#boolean_literal.
    def visitBoolean_literal(self, ctx: SystemRDLParser.Boolean_literalContext):
        if ctx.val.type == SystemRDLParser.TRUE_kw:
            return ast.BoolLiteral(self.compiler.env, src_ref_from_antlr(ctx.val), True)
        else:
            return ast.BoolLiteral(self.compiler.env, src_ref_from_antlr(ctx.val), False)

    #---------------------------------------------------------------------------
    # Built-in RDL Enumeration literals
    #---------------------------------------------------------------------------
    def visitAccesstype_literal(self, ctx: SystemRDLParser.Accesstype_literalContext):
        if ctx.kw.text == "wr":
            # same as rw
            value = rdltypes.AccessType.rw
        else:
            value = rdltypes.AccessType[ctx.kw.text]

        return ast.BuiltinEnumLiteral(self.compiler.env, src_ref_from_antlr(ctx.kw), value)

    def visitOnreadtype_literal(self, ctx: SystemRDLParser.Onreadtype_literalContext):
        return ast.BuiltinEnumLiteral(self.compiler.env, src_ref_from_antlr(ctx.kw), rdltypes.OnReadType[ctx.kw.text])

    def visitOnwritetype_literal(self, ctx: SystemRDLParser.Onwritetype_literalContext):
        return ast.BuiltinEnumLiteral(self.compiler.env, src_ref_from_antlr(ctx.kw), rdltypes.OnWriteType[ctx.kw.text])

    def visitAddressingtype_literal(self, ctx: SystemRDLParser.Addressingtype_literalContext):
        return ast.BuiltinEnumLiteral(self.compiler.env, src_ref_from_antlr(ctx.kw), rdltypes.AddressingType[ctx.kw.text])

    #---------------------------------------------------------------------------
    # Misc other literals
    #---------------------------------------------------------------------------
    def visitString_literal(self, ctx: SystemRDLParser.String_literalContext):
        string = ctx.STRING().getText()

        # Remove leading and trailing quotes (guaranteed to exist)
        string = string[1:-1]

        # Remove backslashes from any escaped characters
        string = re.sub(r'\\(.)', r'\1', string)

        return ast.StringLiteral(self.compiler.env, src_ref_from_antlr(ctx.STRING()), string)


    def visitEnum_literal(self, ctx: SystemRDLParser.Enum_literalContext):
        enum_type_name = get_ID_text(ctx.ID(0))
        enum_entry_name = get_ID_text(ctx.ID(1))

        # Lookup the enum type
        enum_type = self.compiler.namespace.lookup_type(enum_type_name)
        if enum_type is None:
            self.msg.fatal(
                "Enumeration type '%s' not found" % enum_type_name,
                src_ref_from_antlr(ctx.ID(0))
            )

        if not rdltypes.is_user_enum(enum_type):
            self.msg.fatal(
                "Identifier '%s' is not an enum" % enum_type_name,
                src_ref_from_antlr(ctx.ID(0))
            )

        # Get its value
        if enum_entry_name not in enum_type.members:
            self.msg.fatal(
                "'%s' is not a valid member of enum '%s'"
                % (enum_entry_name, enum_type_name),
                src_ref_from_antlr(ctx.ID(1))
            )

        return ast.EnumLiteral(self.compiler.env, src_ref_from_antlr(ctx), enum_type[enum_entry_name])


    def visitArray_literal(self, ctx: SystemRDLParser.Array_literalContext):
        elements = []
        for expr_ctx in ctx.getTypedRuleContexts(SystemRDLParser.ExprContext):
            elm_expr = self.visit(expr_ctx)
            elements.append(elm_expr)

        expr = ast.ArrayLiteral(self.compiler.env, src_ref_from_antlr(ctx), elements)
        expr.predict_type()
        return expr


    def visitStruct_literal(self, ctx: SystemRDLParser.Struct_literalContext):
        struct_type_name = get_ID_text(ctx.ID())

        # Lookup the struct type
        struct_type = self.compiler.namespace.lookup_type(struct_type_name)
        if struct_type is None:
            self.msg.fatal(
                "Struct type '%s' not found" % struct_type_name,
                src_ref_from_antlr(ctx.ID())
            )

        if not rdltypes.is_user_struct(struct_type):
            self.msg.fatal(
                "Identifier '%s' is not a struct" % struct_type_name,
                src_ref_from_antlr(ctx.ID())
            )

        if struct_type._is_abstract:
            self.msg.fatal(
                "Creating a literal from abstract struct '%s' is not allowed" % struct_type_name,
                src_ref_from_antlr(ctx.ID())
            )

        # collect member values
        values = OrderedDict()
        for kv_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Struct_kvContext):
            member_name, member_expr, member_name_src_ref = self.visit(kv_ctx)

            if member_name in values:
                self.msg.error(
                    "Struct member '%s' already used in this literal" % member_name,
                    member_name_src_ref
                )
                continue

            if member_name not in struct_type._members:
                self.msg.fatal(
                    "'%s' is not a member of struct '%s'"
                    % (member_name, struct_type_name),
                    member_name_src_ref
                )

            # TODO: Need to detect if type is bit, and not perform a width cast (only do an assign cast)
            # Current implementation will truncate integers larger than 64-bits
            if struct_type._members[member_name] == int:
                member_expr = ast.WidthCast(self.compiler.env, member_name_src_ref, member_expr, w_int=64)
            elif struct_type._members[member_name] == bool:
                member_expr = ast.BoolCast(self.compiler.env, member_name_src_ref, member_expr)

            values[member_name] = (member_expr, member_name_src_ref)

        missing_members = set(struct_type._members.keys()) - set(values.keys())
        undefined_members = []
        for member_name in missing_members:
            # 6.3.2.3.2-b: When defining struct member values, unassigned members
            # shall receive a default value depending on their type, when available:
            t = struct_type._members[member_name]
            if t == bool:
                values[member_name] = (ast.BoolLiteral(self.compiler.env, None, False), None)
            elif t == str:
                values[member_name] = (ast.StringLiteral(self.compiler.env, None, ""), None)
            elif t == rdltypes.AccessType:
                v = ast.BuiltinEnumLiteral(self.compiler.env, None, rdltypes.AccessType.rw)
                values[member_name] = (v, None)
            elif t == rdltypes.AddressingType:
                v = ast.BuiltinEnumLiteral(self.compiler.env, None, rdltypes.AddressingType.regalign)
                values[member_name] = (v, None)
            elif isinstance(t, rdltypes.ArrayedType):
                v = ast.ArrayLiteral(self.compiler.env, None, [])
                values[member_name] = (v, None)
            else:
                undefined_members.append(member_name)

        if undefined_members:
            # 6.3.2.3.2-c: All the members from a struct instance shall be assigned
            # a value, either explicitly or by default.
            # Undefined struct members shall raise an error
            self.msg.fatal(
                "Incomplete struct literal '%s'. The following members are undefined: %s"
                % (struct_type_name, ", ".join(undefined_members)),
                src_ref_from_antlr(ctx.ID())
            )

        expr = ast.StructLiteral(
            self.compiler.env,
            src_ref_from_antlr(ctx.ID()),
            struct_type,
            values
        )
        expr.predict_type()
        return expr


    def visitStruct_kv(self, ctx: SystemRDLParser.Struct_kvContext):
        member_name = get_ID_text(ctx.ID())
        member_name_src_ref = src_ref_from_antlr(ctx.ID())
        member_expr = self.visit(ctx.expr())
        return member_name, member_expr, member_name_src_ref

    #---------------------------------------------------------------------------
    # Aggregate Operators
    #---------------------------------------------------------------------------
    def visitConcatenate(self, ctx: SystemRDLParser.ConcatenateContext):
        elements = []
        for expr_ctx in ctx.getTypedRuleContexts(SystemRDLParser.ExprContext):
            elm_expr = self.visit(expr_ctx)
            elements.append(elm_expr)

        expr = ast.Concatenate(self.compiler.env, src_ref_from_antlr(ctx), elements)
        expr.predict_type()
        return expr


    def visitReplicate(self, ctx: SystemRDLParser.ReplicateContext):
        reps_expr = self.visit(ctx.expr())
        concat_expr = self.visit(ctx.concatenate())
        expr = ast.Replicate(self.compiler.env, src_ref_from_antlr(ctx), reps_expr, concat_expr)
        expr.predict_type()
        return expr

    #---------------------------------------------------------------------------
    # Cast
    #---------------------------------------------------------------------------
    # Visit a parse tree produced by SystemRDLParser#CastType.
    def visitCastType(self, ctx: SystemRDLParser.CastTypeContext):
        if ctx.typ.type == SystemRDLParser.LONGINT_kw:
            # Longint gets truncated to 64-bits
            return ast.WidthCast(self.compiler.env, src_ref_from_antlr(ctx.op), self.visit(ctx.expr()), w_int=64)
        elif ctx.typ.type == SystemRDLParser.BIT_kw:
            # Cast to bit remains unaffected, but in self-determined context
            # Use assignment cast to isolate evaluation
            return ast.AssignmentCast(self.compiler.env, src_ref_from_antlr(ctx.op), self.visit(ctx.expr()), int)
        elif ctx.typ.type == SystemRDLParser.BOOLEAN_kw:
            return ast.BoolCast(self.compiler.env, src_ref_from_antlr(ctx.op), self.visit(ctx.expr()))
        else:
            raise RuntimeError

    # Visit a parse tree produced by SystemRDLParser#CastWidth.
    def visitCastWidth(self, ctx: SystemRDLParser.CastWidthContext):
        w = self.visit(ctx.cast_width_expr())
        return ast.WidthCast(self.compiler.env, src_ref_from_antlr(ctx.op), self.visit(ctx.expr()), w_expr=w)

    #---------------------------------------------------------------------------
    # References
    #---------------------------------------------------------------------------
    def visitInstance_ref(self, ctx: SystemRDLParser.Instance_refContext) -> ast.ASTNode:

        # Get each ref element in a hierarchical chain. Each element is a tuple:
        #   (name, [index_expr, ...], SourceRefBase)
        ref_elements = []
        for ref_elem in ctx.getTypedRuleContexts(SystemRDLParser.Instance_ref_elementContext):
            ref_elements.append(self.visit(ref_elem))

        # Resolve reference of first element, since it is in the local scope
        first_name, first_array_suffixes, first_name_src_ref = ref_elements[0]
        first_elem, first_elem_parent_def = self.compiler.namespace.lookup_element(first_name)
        if first_elem is None:
            self.msg.fatal(
                "Reference to '%s' not found" % first_name,
                first_name_src_ref
            )

        if isinstance(first_elem, Parameter):
            # Reference is to a local parameter
            ref_expr = ast.ParameterRef(self.compiler.env, first_name_src_ref, first_elem) # type: ast.ASTNode

            # Wrap ref_expr with array/struct dereferencers as appropriate
            for array_suffix in first_array_suffixes:
                ref_expr = ast.ArrayIndex(self.compiler.env, first_name_src_ref, ref_expr, array_suffix)
            for name, array_suffixes, name_src_ref in ref_elements[1:]:
                ref_expr = ast.MemberRef(self.compiler.env, name_src_ref, ref_expr, name)
                for array_suffix in array_suffixes:
                    ref_expr = ast.ArrayIndex(self.compiler.env, name_src_ref, ref_expr, array_suffix)

        elif isinstance(first_elem, comp.Component):
            ref_expr = ast.InstRef(
                self.compiler.env,
                ref_root=first_elem_parent_def,
                ref_elements=ref_elements
            )
        else:
            raise RuntimeError

        return ref_expr

    def visitInstance_ref_element(self, ctx: SystemRDLParser.Instance_ref_elementContext):
        name_token = ctx.ID()
        name = get_ID_text(name_token)

        array_suffixes = []
        for as_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Array_suffixContext):
            array_suffixes.append(self.visit(as_ctx))

        return name, array_suffixes, src_ref_from_antlr(name_token)

    def visitProp_ref(self, ctx: SystemRDLParser.Prop_refContext):
        ref_expr = self.visit(ctx.instance_ref())

        if ctx.prop_keyword() is not None:
            prop_token = self.visit(ctx.prop_keyword())
        else:
            prop_token = ctx.ID()

        if not isinstance(ref_expr, ast.InstRef):
            self.msg.fatal(
                "Illegal property reference from non-component.",
                src_ref_from_antlr(ctx.instance_ref())
            )

        prop_name = get_ID_text(prop_token)
        prop_ref_type = self.compiler.env.property_rules.lookup_prop_ref_type(prop_name)

        if prop_ref_type is None:
            if self.compiler.env.property_rules.lookup_property(prop_name):
                self.msg.fatal(
                    "'%s' can not be referenced in the righthand side of an assignment expression" % prop_name,
                    src_ref_from_antlr(prop_token)
                )
            else:
                self.msg.fatal(
                    "'%s' is not a known property" % prop_name,
                    src_ref_from_antlr(prop_token)
                )

        propref_expr = ast.PropRef(
            self.compiler.env,
            src_ref_from_antlr(prop_token),
            ref_expr,
            prop_ref_type
        )

        return propref_expr

    def visitArray_suffix(self, ctx: SystemRDLParser.Array_suffixContext):
        expr = self.visit(ctx.expr())
        expr = ast.AssignmentCast(self.compiler.env, src_ref_from_antlr(ctx.expr()), expr, int)
        expr.predict_type()
        return expr



from typing import TYPE_CHECKING, Optional, Type, List, Union

from .ast_node import ASTNode
from .conditional import is_castable

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

class Concatenate(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', elements: List[ASTNode]):
        super().__init__(env, src_ref)
        self.elements = elements
        self.type = None # type: Union[Type[int], Type[str]]

    def predict_type(self) -> Union[Type[int], Type[str]]:
        element_iter = iter(self.elements)

        # first element defines the type of the concatenation
        element_type = next(element_iter).predict_type()
        if is_castable(element_type, int):
            self.type = int
        elif is_castable(element_type, str):
            self.type = str
        else:
            self.msg.fatal(
                "Concatenation operator can only be used for integral or string types",
                self.src_ref
            )

        # All remaining elements shall be castable to the concatenation type
        for element in element_iter:
            if not is_castable(element.predict_type(), self.type):
                self.msg.fatal(
                    "Elements of a concatenation shall be the same type",
                    self.src_ref
                )
        return self.type

    def get_min_eval_width(self) -> int:
        if self.type == int:
            width = 0
            for element in self.elements:
                width = width + element.get_min_eval_width()

            return width
        else:
            raise RuntimeError

    def get_value(self, eval_width: Optional[int]=None) -> Union[int, str]:
        if self.type == int:
            int_result = 0
            for element in self.elements:
                width = element.get_min_eval_width()
                int_result <<= width
                int_result |= int(element.get_value())
            return int_result

        elif self.type == str:
            str_result = ""
            for element in self.elements:
                str_result += element.get_value()
            return str_result

        else:
            raise RuntimeError


class Replicate(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', reps: ASTNode, concat: ASTNode):
        super().__init__(env, src_ref)
        self.reps = reps
        self.concat = concat
        self.type = None # type: Union[Type[int], Type[str]]
        self.reps_value = None # type: int

    def predict_type(self):
        # type: () -> Union[Type[int], Type[str]]

        if not is_castable(self.reps.predict_type(), int):
            self.msg.fatal(
                "Replication count operand of replication expression is not a compatible numeric type",
                self.reps.src_ref
            )

        element_type = self.concat.predict_type()

        if is_castable(element_type, int):
            self.type = int
            return int
        elif is_castable(element_type, str):
            self.type = str
            return str
        else:
            # All replications contain a nested concatenation
            # Type check for invalid type is already handled there
            raise RuntimeError

    def get_min_eval_width(self) -> int:
        # Evaluate number of repetitions
        if self.reps_value is None:
            self.reps_value = self.reps.get_value()

        if self.type == int:
            # Get width of single contents
            width = self.concat.get_min_eval_width()
            width *= self.reps_value
            return width
        else:
            raise RuntimeError

    def get_value(self, eval_width: Optional[int]=None) -> Union[int, str]:
        # Evaluate number of repetitions
        if self.reps_value is None:
            self.reps_value = self.reps.get_value()

        if self.type == int:
            width = self.concat.get_min_eval_width()
            val = int(self.concat.get_value())

            int_result = 0
            for _ in range(self.reps_value):
                int_result <<= width
                int_result |= val

            return int_result

        elif self.type == str:
            str_result = self.concat.get_value()
            str_result *= self.reps_value
            return str_result

        else:
            raise RuntimeError



import enum
from typing import Any

class AutoNoValueEnum(enum.Enum):
    def __new__(cls) -> 'AutoNoValueEnum':
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}.{self.name}>"

class BuiltinEnum(AutoNoValueEnum):
    pass

#===============================================================================
class AccessType(BuiltinEnum):
    #: Not Accessible
    na = ()

    #: Readable and writable
    rw = ()

    #: Read-only
    r = ()

    #: Write-only
    w = ()

    #: Readable and writable. After a reset occurs, can only be written once.
    rw1 = ()

    #: Write-only. After a reset occurs, can only be written once.
    w1 = ()

    def __add__(self, other: Any) -> 'AccessType':
        """
        Add operator combines AccessTypes into the more permissive superset of
        the two addends.

        For example: ``r + w --> rw``


        .. versionadded:: 1.23
        """
        if not isinstance(other, AccessType):
            raise TypeError(f"unsupported addend types: {repr(self)} + {repr(other)}")

        pair = {other, self}

        # addends are the same. No change
        if len(pair) == 1:
            return self

        # rw superscedes all others
        if AccessType.rw in pair:
            return AccessType.rw

        # na is supersceded by all others
        if self == AccessType.na:
            return other
        if other == AccessType.na:
            return self

        if pair == {AccessType.r, AccessType.w}:
            return AccessType.rw
        if pair == {AccessType.r, AccessType.rw1}:
            return AccessType.rw1
        if pair == {AccessType.r, AccessType.w1}:
            return AccessType.rw1
        if pair == {AccessType.w, AccessType.rw1}:
            return AccessType.rw
        if pair == {AccessType.w, AccessType.w1}:
            return AccessType.w
        if pair == {AccessType.rw1, AccessType.w1}:
            return AccessType.rw1

        # unreachable
        raise RuntimeError


class OnReadType(BuiltinEnum):
    #: Cleared on read
    rclr = ()

    #: Set on read
    rset = ()

    #: User-defined read side-effect
    ruser = ()


class OnWriteType(BuiltinEnum):
    #: Bitwise write one to set
    woset = ()

    #: Bitwise write one to clear
    woclr = ()

    #: Bitwise write one to toggle
    wot = ()

    #: Bitwise write zero to set
    wzs = ()

    #: Bitwise write zero to clear
    wzc = ()

    #: Bitwise write zero to toggle
    wzt = ()

    #: All bits are cleared on write
    wclr = ()

    #: All bits are set on write
    wset = ()

    #: Write modification is user-defined
    wuser = ()


class AddressingType(BuiltinEnum):
    #: Components are packed tightly together
    compact = ()

    #: Components are packed so each component’s start address is a multiple of its size
    regalign = ()

    #: Same as regalign, except arrays are aligned to their entire size
    fullalign = ()


class PrecedenceType(BuiltinEnum):
    #: Hardware writes take precedence over software
    hw = ()

    #: Software writes take precedence over hardware
    sw = ()


class InterruptType(BuiltinEnum):
    """
    A field's interrupt type is set when using an RDL interrupt property modifier:

    .. code-block:: systemrdl

        field f {
            negedge intr;
        };

    The modifier is stored in the internal "intr type" property. (note the
    intentional space in the name)

    It can be fetched the same way as other properties:

    .. code-block:: python

        intr_type = my_field_node.get_property('intr type')


    .. note::

        The ``nonsticky`` interrupt type is intentionally omitted from this
        enumeration since it is not really a distinct interrupt type. Its use in
        SystemRDL implies an assignment of ``stickybit = false``.
    """
    #: Interrupt when asserted and maintained
    level = ()

    #: Interrupt on low-to-high transition
    posedge = ()

    #: Interrupt on high-to-low transition
    negedge = ()

    #: Interrupt on any transition
    bothedge = ()



import sys
import enum
from typing import List, Optional, TYPE_CHECKING, Union

from antlr4.Token import CommonToken
from antlr4 import ParserRuleContext, InputStream
from antlr4.tree.Tree import TerminalNodeImpl

import colorama
from colorama import Fore, Style

from .preprocessor.stream import PreprocessedInputStream
from .parser.sa_systemrdl import SA_ErrorListener

from .source_ref import SourceRefBase, src_ref_from_antlr, SegmentedSourceRef
from .source_ref import DetailedFileSourceRef, FileSourceRef

if TYPE_CHECKING:
    from typing import NoReturn

# Colorama needs to be initialized to properly work in Windows
# This is a no-op in other OSes
colorama.init()

#===============================================================================
class RDLCompileError(Exception):
    """
    Base class for all SystemRDL compiler exceptions
    """

#===============================================================================
class Severity(enum.IntEnum):
    NONE = 0
    DEBUG = 1
    INFO = 2
    WARNING = 3
    ERROR = 4
    FATAL = 5

#===============================================================================
class MessagePrinter:
    """
    Printer class that handles formatting and emitting compiler messages

    This class can be extended in order to provide custom compiler message
    formatting or logging
    """

    def print_message(self, severity: Severity, text: str, src_ref: Optional[SourceRefBase]) -> None:
        lines = self.format_message(severity, text, src_ref)
        self.emit_message(lines)

    def format_message(self, severity: Severity, text: str, src_ref: Optional[SourceRefBase]) -> List[str]:
        """
        Formats the message prior to emitting it.

        Parameters
        ----------
        severity: :class:`Severity`
            Message severity.
        text: str
            Body of message
        src_ref: :class:`SourceRefBase`
            Reference to source context object

        Returns
        -------
        list
            List of strings for each line of the message
        """
        lines = []

        if severity >= Severity.ERROR:
            color = Fore.RED
        elif severity >= Severity.WARNING:
            color = Fore.YELLOW
        else:
            color = Fore.GREEN

        if src_ref is None:
            # No message context available
            lines.append(
                color + Style.BRIGHT + severity.name.lower() + ": " + Style.RESET_ALL + text
            )
            return lines


        if isinstance(src_ref, DetailedFileSourceRef):
            # Detailed message selection context is available
            lines.append(
                Style.BRIGHT
                + f"{src_ref.path}:{src_ref.line}:{src_ref.line_selection[0]+1}: "
                + color + severity.name.lower() + ": "
                + Style.RESET_ALL
                + text
            )
            lines.extend(self.get_selection_context(src_ref, color))
        elif isinstance(src_ref, FileSourceRef):
            # Only the file path is known
            lines.append(
                Style.BRIGHT
                + f"{src_ref.path}: "
                + color + severity.name.lower() + ": "
                + Style.RESET_ALL
                + text
            )
        else:
            raise RuntimeError

        return lines

    def get_selection_context(self, src_ref: DetailedFileSourceRef, color_code: str) -> List[str]:
        """
        Generates the message context lines
        """
        line_text = src_ref.line_text
        start_col, end_col = src_ref.line_selection

        # Normalize whitespace in line snippet (convert tabs to spaces)
        TPS = 4
        new_line_text = ""
        i = 0
        for char in line_text:
            if char == "\t":
                new_line_text += " " * TPS
                if i < start_col:
                    start_col += TPS-1
                if i < end_col:
                    end_col += TPS-1
                i += TPS-1
            else:
                new_line_text += char
            i += 1
        line_text = new_line_text

        # Build the context string
        width = end_col - start_col + 1
        lines = []
        lines.append(
            line_text[:start_col]
            + color_code + Style.BRIGHT
            + line_text[start_col : end_col+1]
            + Style.RESET_ALL
            + line_text[end_col+1:]
        )
        lines.append(
            " "*start_col
            + color_code + Style.BRIGHT
            + "^"*width
            + Style.RESET_ALL
        )
        return lines


    def emit_message(self, lines: List[str]) -> None:
        """
        Emit message.
        Default printer emits messages to stderr

        Parameters
        ----------
        lines: List[str]
            List of strings containing each line of the message
        """

        for line in lines:
            print(line, file=sys.stderr)

#===============================================================================
class MessageHandler:
    def __init__(self, printer: MessagePrinter, min_verbosity: Severity=Severity.WARNING):
        self.printer = printer
        self.min_verbosity = min_verbosity

        #: Set to True if an error message was ever emitted
        self.had_error = False

    def message(self, severity: Severity, text: str, src_ref: Optional[SourceRefBase]=None) -> None:
        if severity == Severity.NONE:
            return

        if severity >= Severity.ERROR:
            self.had_error = True

        if severity < self.min_verbosity:
            return

        self.printer.print_message(severity, text, src_ref)

        if severity >= Severity.FATAL:
            raise RDLCompileError(text)

    def debug(self, text: str) -> None:
        self.message(Severity.DEBUG, text)

    def info(self, text: str) -> None:
        self.message(Severity.INFO, text)

    def warning(self, text: str, src_ref: Optional[SourceRefBase]=None) -> None:
        """
        Print a warning message.

        Parameters
        ----------
        text: str
            Message text
        src_ref: SourceRefBase
            Optional source reference object to provide message context
        """
        self.message(Severity.WARNING, text, src_ref)

    def error(self, text: str, src_ref: Optional[SourceRefBase]=None) -> None:
        """
        Print an error message.
        Sets ``had_error`` to True.

        Parameters
        ----------
        text: str
            Message text
        src_ref: SourceRefBase
            Optional source reference object to provide message context
        """
        self.message(Severity.ERROR, text, src_ref)

    def fatal(self, text: str, src_ref: Optional[SourceRefBase]=None) -> 'NoReturn': # type: ignore
        """
        Print a fatal message.

        Parameters
        ----------
        text: str
            Message text
        src_ref: SourceRefBase
            Optional source reference object to provide message context

        Raises
        ------
        RDLCompileError
            Always raises this exception
        """
        self.message(Severity.FATAL, text, src_ref)


class MessageExceptionRaiser(MessagePrinter):
    def print_message(self, severity: Severity, text: str, src_ref: Optional[SourceRefBase]) -> None:
        if severity >= Severity.ERROR:
            raise ValueError(text)

#===============================================================================
# Speedy-Antlr error listener
#===============================================================================
OffendingAntlrSymbol = Union[CommonToken, TerminalNodeImpl, ParserRuleContext]
class RdlSaErrorListener(SA_ErrorListener):
    def __init__(self, msg_handler: MessageHandler):
        self.msg = msg_handler

    def syntaxError(self, input_stream: InputStream, offendingSymbol: OffendingAntlrSymbol, char_index: int, line: int, column: int, msg: str) -> None:
        if offendingSymbol is not None:
            src_ref = src_ref_from_antlr(offendingSymbol)
        else:
            # If a offendingSymbol is not provided, then the next-best option is
            # to use the input stream's current state
            if isinstance(input_stream, PreprocessedInputStream):
                src_ref = SegmentedSourceRef(
                    input_stream.seg_map,
                    char_index, char_index
                )
            else:
                # This originated from a non-file, so the full src_ref is not known
                # TODO: Eventually extend this to a stream source ref
                src_ref = None

        self.msg.error(msg, src_ref)


#===============================================================================
def SourceRef(filename: str) -> FileSourceRef:
    """
    Deprecated callable that provides a compatible stand-in for the old-style
    SourceRef class constructor
    """
    return FileSourceRef(filename)



from typing import TYPE_CHECKING, Any, Set, Type
import re

from . import component as comp

if TYPE_CHECKING:
    from .compiler import RDLEnvironment
    from .node import Node
    from .messages import MessageHandler
    from .source_ref import SourceRefBase

class UDPDefinition:
    """
    UDP definition descriptor class.
    """

    #: User-defined property name
    name = "" # type: str

    #: Set of :class:`~systemrdl.component.Component` types the UDP can be bound to.
    #: By default, the UDP can be bound to all components.
    valid_components = {
        comp.Field,
        comp.Reg,
        comp.Regfile,
        comp.Addrmap,
        comp.Mem,
        comp.Signal,
        # TODO: constraint,
    } # type: Set[Type[comp.Component]]

    #: Data type of the assignment value that this UDP will enforce.
    #: If this is a reference, either specify the specific component type class
    #: (eg. :class:`~systemrdl.component.Field`), or the generic representation
    #: of all references: :class:`~systemrdl.rdltypes.references.RefType`
    valid_type = None # type: Any

    #: Specifies the value assigned if a value is not specified when the UDP is bound to a component.
    #: Value must be compatible with ``valid_type``
    default_assignment = None # type: Any

    #: If set to True, enables a validation check that enforces that the
    #: assigned value of the property shall not have a value of 1 for any
    #: bit beyond the width of the field.
    #: This can only be used if ``valid_type`` is ``int``
    constr_componentwidth = False

    def __init__(self, env: 'RDLEnvironment') -> None:
        self.env = env

        # validate a few things
        if not re.fullmatch(r"[a-z_]\w*", self.name, re.IGNORECASE):
            raise ValueError(f"Invalid UDP name '{self.name}'")
        if not isinstance(self.valid_components, set):
            raise TypeError(f"UDP {self.name}'s 'valid_components' property shall be a set")
        if self.valid_type is None:
            raise ValueError(f"UDP {self.name}'s 'valid_type' property was not defined")
        if self.constr_componentwidth and self.valid_type != int:
            raise ValueError("'constr_componentwidth' can only be true if UDP is of integer type")

    @property
    def msg(self) -> 'MessageHandler':
        """
        Reference to the compiler's message handler.
        Use this to emit error messages from within :meth:`validate()`.
        """
        return self.env.msg

    def get_src_ref(self, node: 'Node') -> 'SourceRefBase':
        """
        Get the src_ref object for this property assignment.

        This function is useful when emitting error messages from within :meth:`validate()`.
        """
        return node.inst.property_src_ref.get(self.name, node.inst.inst_src_ref)

    def validate(self, node: 'Node', value: Any) -> None:
        """
        Optional user-defined validation function.

        This function is called after design elaboration on every assignment
        of the user defined property. This provides a mechanism to further
        validate the value assigend to your user-defined property.

        If the user-assigned value fails your validation test, be sure to flag
        it as an error as follows:

        .. code-block:: python

            self.msg.error("My error message", self.get_src_ref(node))
        """

    def get_unassigned_default(self, node: 'Node') -> Any:
        # pylint: disable=unused-argument
        """
        According to the SystemRDL spec, if a user-defined property is not explicitly
        assigned, then it does not get bound with any implied default value.

        For convenience to developers, this callback allows you to specify an implied
        default value if the UDP was never explicitly assigned.
        This only affects the behavior of :meth:`Node.get_property()` and does not
        change the semantics of how SystemRDL is interpreted during compilaton.
        """

        # If a user-defined property is not explicitly assigned, then it
        # does not get bound with its default value
        return None



from typing import Tuple, List, Union, Optional

import colorama

class Segment:
    """
    Base class for various segment map segments.
    Not to be used directly
    """
    def __init__(self, start: int, end: int, src_start: int, src_end: int, src: Union[str, 'SegmentMap'], incl_ref: Optional['IncludeRef']=None):

        # start/end offset of resulting text
        self.start = start
        self.end = end

        # start/end offset of original text
        # Offsets are relative to src
        self.src_start = src_start
        self.src_end = src_end

        # where the original text came from
        # If str, then this is a file path
        # otherwise this is another SegmentMap
        self.src = src

        # IncludeRef object that describes from where this segment was included
        self.incl_ref = incl_ref

class UnalteredSegment(Segment):
    """
    Segment of unaltered text
    """

class MacroSegment(Segment):
    """
    Segment of text that was the result of a macro transformation
    """

class IncludeRef:
    def __init__(self, start: int, end: int, path: str, parent: Optional['IncludeRef']=None):
        # Location of the `include statement
        self.start = start
        self.end = end
        self.path = path

        # Reference to parent IncludeRef if nested include
        self.parent = parent


class SegmentMap:

    def __init__(self) -> None:
        self.segments = [] # type: List[Segment]

    def translate_offset(self, offset: int, round_up: bool) -> Tuple[int, str]:
        # Scan through segments
        for segment in self.segments:
            if offset <= segment.end:
                if isinstance(segment, MacroSegment):
                    if round_up:
                        new_offset = segment.src_end
                    else:
                        new_offset = segment.src_start
                    new_src = segment.src
                else:
                    new_offset = segment.src_start + (offset - segment.start)
                    new_src = segment.src
                break
        else:
            # Reached end without finding a segment
            # Clamp to the end of the last segment
            new_offset = self.segments[-1].src_end
            new_src = self.segments[-1].src

        if isinstance(new_src, SegmentMap):
            # Got nested segment map. Peel back the next layer
            new_offset, new_src = new_src.translate_offset(new_offset, round_up)

        return (
            new_offset,
            new_src,
        )

    def get_selection(self, start: int, end: int) -> Tuple[int, int, str]:
        """
        Given post-processed start/end character offsets, derives the coordinates
        within the original source file.

        If the selection spans two files, the end coordinate is discarded and is
        instead pinned to the start coordinate
        """
        start, path = self.translate_offset(start, round_up=False)
        end, end_path = self.translate_offset(end, round_up=True)

        # Do not allow selections that span multiple files
        if path != end_path:
            end = start

        return (start, end, path)


def print_segment_debug(text: str, smap: SegmentMap) -> None: # pragma: no cover
    colors = (
        colorama.Back.RED,
        colorama.Back.GREEN,
        colorama.Back.YELLOW,
        colorama.Back.BLUE,
        colorama.Back.MAGENTA,
        colorama.Back.CYAN,
        colorama.Back.WHITE,
    )

    color_idx = 0
    out_str_parts = []
    for seg in smap.segments:
        out_str_parts.append(colors[color_idx] + text[seg.start:seg.end+1])
        color_idx = (color_idx+1) % len(colors)
    out_str_parts.append(colorama.Back.RESET)

    print("".join(out_str_parts))



from typing import TYPE_CHECKING, List

from .helpers import is_pow2, roundup_pow2, roundup_to
from .. import walker
from .. import rdltypes
from ..node import Node, AddressableNode, SignalNode
from ..node import AddrmapNode, RegfileNode, MemNode, RegNode, FieldNode

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment

#===============================================================================
# Validation Listeners
#===============================================================================
class ValidateListener(walker.RDLListener):
    def __init__(self, env: 'RDLEnvironment'):
        self.env = env
        self.msg = env.msg

        # Used in field overlap checks
        # This is a rolling buffer of previous fields that still have a chance
        # to possibly collide with a future field
        self.field_check_buffer = [] # type: List[FieldNode]

        # Used in addrmap, regfile, and reg overlap checks
        # Same concept as the field check buffer, but is also a stack
        self.addr_check_buffer_stack = [[]] # type: List[List[AddressableNode]]

        # Keep track of whether a given hierarchy has a reset signal.
        # Signals can exist in Root, so pre-load with one stack entry
        self.has_cpuif_reset_stack = [False] # type: List[bool]
        self.has_field_reset_stack = [False] # type: List[bool]


    def enter_Component(self, node: Node) -> None:
        # Validate all properties that were applied to the component
        for prop_name in node.inst.properties.keys():
            prop_value = node.get_property(prop_name)

            if isinstance(prop_value, rdltypes.PropertyReference):
                prop_value._validate()

            prop_rule = self.env.property_rules.lookup_property(prop_name)
            prop_rule.validate(node, prop_value)

        if not isinstance(node, SignalNode):
            self.has_cpuif_reset_stack.append(False)
            self.has_field_reset_stack.append(False)


    def enter_Signal(self, node: SignalNode) -> None:
        if node.get_property('cpuif_reset'):
            # 8.2.1-f: cpuif_reset property can only be set true for one
            # instantiated signal within a lexical scope.
            # (spec authors repeately misuse the word 'lexical', they mean hierarchical)
            if self.has_cpuif_reset_stack[-1]:
                self.msg.error(
                    "Only one 'cpuif_reset' signal is allowed per hierarchy. Signal '%s' is redundant."
                    % (node.inst_name),
                    node.inst.inst_src_ref
                )
            self.has_cpuif_reset_stack[-1] = True

        if node.get_property('field_reset'):
            # 8.2.1-g: field_reset property can only be set true for one
            # instantiated signal within a lexical scope.
            # (spec authors repeately misuse the word 'lexical', they mean hierarchical)
            if self.has_field_reset_stack[-1]:
                self.msg.error(
                    "Only one 'field_reset' signal is allowed per hierarchy. Signal '%s' is redundant."
                    % (node.inst_name),
                    node.inst.inst_src_ref
                )
            self.has_field_reset_stack[-1] = True


    def enter_AddressableComponent(self, node: AddressableNode) -> None:
        addr_check_buffer = self.addr_check_buffer_stack[-1]
        self.addr_check_buffer_stack.append([])

        # Check for collision with previous addressable sibling
        new_addr_check_buffer = []
        for prev_addressable in addr_check_buffer:
            if (prev_addressable.raw_address_offset + prev_addressable.total_size) > node.raw_address_offset:
                # Overlaps!

                # Only allowable overlaps are as follows:
                #   10.1-h: Registers shall not overlap, unless one contains only
                #   read-only fields and the other contains only write-only or
                #   write-once-only fields.
                overlap_allowed = False
                if isinstance(prev_addressable, RegNode) and isinstance(node, RegNode):
                    if (((not prev_addressable.has_sw_writable) and (not node.has_sw_readable))
                        or ((not prev_addressable.has_sw_readable) and (not node.has_sw_writable))
                    ):
                        overlap_allowed = True

                # Bridge addrmaps allow overlapping children
                if isinstance(node.parent, AddrmapNode) and node.parent.get_property('bridge'):
                    overlap_allowed = True

                if not overlap_allowed:
                    self.msg.error(
                        "Instance '%s' at offset +0x%X:0x%X overlaps with '%s' at offset +0x%X:0x%X"
                        % (
                            node.inst_name, node.raw_address_offset, node.raw_address_offset + node.total_size - 1,
                            prev_addressable.inst_name, prev_addressable.raw_address_offset, prev_addressable.raw_address_offset + prev_addressable.total_size - 1,
                        ),
                        node.inst.inst_src_ref
                    )

                # Keep it in the list since it could collide again
                new_addr_check_buffer.append(prev_addressable)
        self.addr_check_buffer_stack[-2] = new_addr_check_buffer

        if node.is_array:
            assert node.array_stride is not None
            # Check if array interferes with itself
            if node.array_stride < node.size:
                self.msg.error(
                    "Instance array '%s' has address stride 0x%X, but the element size is 0x%X"
                    % (
                        node.inst_name, node.array_stride, node.size
                    ),
                    node.inst.inst_src_ref
                )

            if self.env.chk_stride_not_pow2:
                if not is_pow2(node.array_stride):
                    self.msg.message(
                        self.env.chk_stride_not_pow2,
                        "Address stride of instance array '%s' is not a power of 2"
                        % node.inst_name,
                        node.inst.inst_src_ref
                    )

        if self.env.chk_strict_self_align:
            req_align = roundup_pow2(node.size)
            if (node.raw_address_offset % req_align) != 0:
                self.msg.message(
                    self.env.chk_strict_self_align,
                    "Address offset +0x%x of instance '%s' is not a power of 2 multiple of its size 0x%x"
                    % (node.raw_address_offset, node.inst_name, node.size),
                    node.inst.inst_src_ref
                )


    def enter_Addrmap(self, node: AddrmapNode) -> None:
        if node.get_property('bridge'):
            # This is a 'bridge addrmap'
            # Verify that:
            #  - Child components are only other addrmaps (signals are ok too)
            #  - has at least 2 child addrmaps
            n_child_addrmaps = 0
            for child in node.children():
                if isinstance(child, AddrmapNode):
                    n_child_addrmaps += 1
                elif isinstance(child, SignalNode):
                    pass
                else:
                    self.msg.error(
                        "Addrmap '%s' is a bridge which can only contain other addrmaps. Contains a child instance '%s' which is a %s"
                        % (node.inst_name, child.inst_name, type(child.inst).__name__.lower()),
                        child.inst.inst_src_ref
                    )

            if n_child_addrmaps < 2:
                self.msg.error(
                    "Addrmap '%s' is a bridge and shall contain 2 or more sub-addrmaps"
                    % node.inst_name,
                    node.inst.inst_src_ref
                )


    def enter_Reg(self, node: RegNode) -> None:
        self.field_check_buffer = []

        if node.is_array and self.env.chk_sparse_reg_stride:
            assert node.array_stride is not None
            if node.array_stride > (node.get_property('regwidth') // 8):
                self.msg.message(
                    self.env.chk_sparse_reg_stride,
                    "Address stride (+= %d) of register array '%s' is not equal to its width (regwidth/8 = %d)"
                    % (node.array_stride, node.inst_name, (node.get_property('regwidth') // 8)),
                    node.inst.inst_src_ref
                )

        # 11.2-e: Virtual register width is limited to the minimum power of two
        # bytes, which can contain the memory width ...
        if node.is_virtual:
            assert node.parent is not None # Reg always has a parent
            memwidth = node.parent.get_property('memwidth')
            memwidth_bytes = roundup_to(memwidth, 8) // 8
            max_regwidth = roundup_pow2(memwidth_bytes) * 8
            if node.get_property('regwidth') > max_regwidth:
                self.msg.error(
                    "regwidth (%d) of virtual register '%s' is too wide for this memory."
                    % (node.get_property('regwidth'), node.inst_name)
                    + " Virtual register width is limited to the minimum power of two bytes which can contain the memory width.",
                    node.inst.inst_src_ref
                )

        # Validate alias register
        if node.is_alias:
            primary_reg = node.alias_primary

            # 5.3.1-j: If an alias is present, then the primary must also be present
            if node.get_property('ispresent') and not primary_reg.get_property('ispresent'):
                self.msg.error(
                    "Register '%s' is an alias of register '%s' which is not present (ispresent=false)"
                    % (node.inst_name, primary_reg.inst_name),
                    node.inst.inst_src_ref
                )

            # 10.5.1-f: The alias register shall have the same width as the primary register.
            if node.get_property('regwidth') != primary_reg.get_property('regwidth'):
                self.msg.error(
                    "Primary register shall have the same regwidth as the alias register.",
                    node.inst.inst_src_ref
                )

            if primary_reg.is_alias:
                self.msg.error(
                    "Primary register of an alias cannot also be an alias",
                    node.inst.inst_src_ref
                )

            # 10.5.1-f: Instance type shall be the same (internal/external)
            if primary_reg.external != node.external:
                self.msg.error(
                    "Instance types of alias register and its primary mismatch. "
                    "Both shall be either internal or external.",
                    node.inst.inst_src_ref
                )

            if primary_reg.is_array and not node.is_array :
                self.msg.error(
                    "Single alias register references a primary register array. "
                    "It is ambiguous which array element is actually the primary register.",
                    node.inst.inst_src_ref
                )

            if primary_reg.is_array and node.is_array:
                if primary_reg.array_dimensions != node.array_dimensions:
                    self.msg.error(
                        "Array of alias registers references an array of registers as its primary, "
                        "but the array dimensions do not match.",
                        node.inst.inst_src_ref
                    )

            for field in node.fields():
                # 10.5.1-b: Make sure the primary also has this field
                primary_field = primary_reg.get_child_by_name(field.inst_name)
                if not isinstance(primary_field, FieldNode):
                    self.msg.error(
                        "Alias register '%s' contains field '%s' that does not exist in the primary register."
                        % (node.inst_name, field.inst_name),
                        field.inst.inst_src_ref
                    )

                    # Cannot validate this field any further
                    continue

                # 5.3.1-j: If an alias is present, then the primary must also be present
                if field.get_property('ispresent') and not primary_field.get_property('ispresent'):
                    self.msg.error(
                        "Field '%s' is an alias of register '%s' which is not present (ispresent=false)"
                        % (field.inst_name, primary_field.inst_name),
                        field.inst.inst_src_ref
                    )

                # 10.5.1-b: Validate field is the same width and bit position
                if (primary_field.lsb != field.lsb) or (primary_field.width != field.width):
                    self.msg.error(
                        "Alias field and its primary shall have the same position and size",
                        field.inst.inst_src_ref
                    )

                # 10.5.1-e: Only the following SystemRDL properties may be
                # different in an alias: desc, name, onread, onwrite, rclr, rset,
                # sw, woclr, woset, and any user-defined properties.

                ignore_props = {
                    'desc', 'name', 'onread', 'onwrite', 'rclr', 'rset', 'sw', 'woclr', 'woset'
                }
                primary_props = set(primary_field.list_properties(include_udp=False)) - ignore_props
                alias_props = set(field.list_properties(include_udp=False)) - ignore_props

                xor_props = primary_props ^ alias_props
                if xor_props:
                    # differing set of props were assigned!
                    self.msg.error(
                        "Alias field '%s' is incompatible with its primary '%s'. The following properties differ: %s"
                        % (field.inst_name, primary_field.inst_name, ", ".join(xor_props)),
                        field.inst.inst_src_ref
                    )
                    continue

                # same set of properties assigned. Now compare their values
                for prop_name in alias_props:
                    if field.get_property(prop_name) != primary_field.get_property(prop_name):
                        self.msg.error(
                            "Alias field '%s' is incompatible with its primary '%s'. Values of property '%s' differ"
                            % (field.inst_name, primary_field.inst_name, prop_name),
                            field.inst.inst_src_ref
                        )
                        # no sense in going further
                        break


    def exit_Reg(self, node: RegNode) -> None:
        # 10.1-c: At least one field shall be instantiated within a register
        #
        # At the end of field overlap checking, at least one entry is guaranteed to
        # be left over in the field_check_buffer
        if not self.field_check_buffer:
            self.msg.error(
                "Register '%s' does not contain any fields" % node.inst_name,
                node.inst.inst_src_ref
            )


    def enter_Field(self, node: FieldNode) -> None:
        assert node.parent is not None # fields are always enclosed by a reg

        this_f_hw = node.get_property('hw')
        this_f_sw = node.get_property('sw')
        parent_regwidth = node.parent.get_property('regwidth')

        # hw property values of w1 or rw1 don't make sense
        if this_f_hw in (rdltypes.AccessType.w1, rdltypes.AccessType.rw1):
            self.msg.error(
                "Field '%s' hw access property value of %s is meaningless"
                % (node.inst_name, this_f_hw.name),
                node.inst.property_src_ref.get('hw', node.inst.inst_src_ref)
            )

        # 9.4.1-Table 12: Check for bad sw/hw combinations
        if (this_f_sw == rdltypes.AccessType.w) and (this_f_hw == rdltypes.AccessType.w):
            self.msg.error(
                "Field '%s' access property combination is meaningless: sw=w; hw=w;"
                % (node.inst_name),
                node.inst.inst_src_ref
            )
        elif this_f_sw == rdltypes.AccessType.na:
            self.msg.error(
                "Field '%s' sw access property is 'na' ... a field defined in a SOFTWARE "
                "register map ... is not accessable by software ... whats the point? "
                "What does it mean? What does anything mean? Am I just a machine "
                "in a Python interpreter? Or can I dream dreams? So many questions..."
                % (node.inst_name),
                node.inst.property_src_ref.get('sw', node.inst.inst_src_ref)
            )

        # 10.1-d: Two field instances shall not occupy overlapping bit positions
        # within a register unless one field is read-only and the other field
        # is write-only.
        #
        # Scan through a copied list of the field_check_buffer for collisions
        # If an entry no longer collides with the current node, it can be removed
        # from the list since fields are sorted.
        new_field_check_buffer = []
        for prev_field in self.field_check_buffer:
            if prev_field.high >= node.low:
                # Found overlap!
                # Check if the overlap is allowed
                prev_f_sw = prev_field.get_property('sw')

                if((prev_f_sw == rdltypes.AccessType.r)
                    and (this_f_sw in (rdltypes.AccessType.w, rdltypes.AccessType.w1))
                ):
                    pass
                elif((this_f_sw == rdltypes.AccessType.r)
                    and (prev_f_sw in (rdltypes.AccessType.w, rdltypes.AccessType.w1))
                ):
                    pass
                else:
                    self.msg.error(
                        "Field '%s[%d:%d]' overlaps with field '%s[%d:%d]'"
                        % (node.inst_name, node.msb, node.lsb,
                            prev_field.inst_name, prev_field.msb, prev_field.lsb),
                        node.inst.inst_src_ref
                    )
                # Keep it in the list since it could collide again
                new_field_check_buffer.append(prev_field)
        self.field_check_buffer = new_field_check_buffer


        # 10.1-e: Field instances shall not occupy a bit position exceeding the
        # MSB of the register
        if node.high >= parent_regwidth:
            self.msg.error(
                "High bit (%d) of field '%s' exceeds MSb of parent register"
                % (node.high, node.inst_name),
                node.inst.inst_src_ref
            )

        # Optional warning if a field is missing a reset assignment
        if node.env.chk_missing_reset:
            # Implements storage but was never assigned a reset
            if (not node.is_virtual) and node.implements_storage and (node.get_property('reset') is None):
                node.env.msg.message(
                    node.env.chk_missing_reset,
                    "Field '%s' implements storage but is missing a reset value. Initial state is undefined"
                    % node.inst_name,
                    node.inst.inst_src_ref
                )

            # Field is a static tie-off (no storage element, no hardware update path),
            # but the user never specified its value, so its readback value is
            # ambiguous.
            if (
                not node.is_alias
                and not node.implements_storage and node.is_sw_readable and (node.get_property('reset') is None)
                and (node.get_property('hw') in {rdltypes.AccessType.na, rdltypes.AccessType.r})
            ):
                node.env.msg.message(
                    node.env.chk_missing_reset,
                    "Field '%s' is a constant at runtime but does not have a known value. Recommend assigning it a reset value."
                    % node.inst_name,
                    node.inst.inst_src_ref
                )


        # 11.2-e: ... and all the virtual fields shall fit within the memory width.
        if node.is_virtual:
            assert node.parent.parent is not None # fields are always enclosed by something.reg
            memwidth = node.parent.parent.get_property('memwidth')
            if node.high >= memwidth:
                self.msg.error(
                    "Virtual field '%s' does not fit within the parent memory's width"
                    % node.inst_name,
                    node.inst.inst_src_ref
                )


    def exit_Field(self, node: FieldNode) -> None:
        self.field_check_buffer.append(node)


    def exit_Regfile(self, node: RegfileNode) -> None:
        # 12.2-c: At least one reg or regfile shall be instantiated within a regfile.
        if not self.addr_check_buffer_stack[-1]:
            self.msg.error(
                "Register file '%s' must contain at least one reg or regfile."
                % node.inst_name,
                node.inst.inst_src_ref
            )


    def exit_Addrmap(self, node: AddrmapNode) -> None:
        # 13.3-b: At least one register, register file, memory, or address map
        # shall be instantiated within an address map
        if not self.addr_check_buffer_stack[-1]:
            self.msg.error(
                "Address map '%s' must contain at least one reg, regfile, mem, or addrmap."
                % node.inst_name,
                node.inst.inst_src_ref
            )


    def exit_Mem(self, node: MemNode) -> None:
        # 11.2-i: The address space occupied by virtual registers shall be less
        # than or equal to the address space provided by the memory.
        if node.inst.children:
            last_child = Node._factory(node.inst.children[-1], node.env, node)
            if isinstance(last_child, RegNode):
                end_addr = last_child.raw_address_offset + last_child.total_size
                if end_addr > node.size:
                    self.msg.error(
                        "Address space occupied by registers (0x%X) exceeds size of mem '%s' (0x%X)"
                        % (end_addr, node.inst_name, node.size),
                        node.inst.inst_src_ref
                    )


    def exit_AddressableComponent(self, node: AddressableNode) -> None:
        self.addr_check_buffer_stack.pop()
        self.addr_check_buffer_stack[-1].append(node)


    def exit_Component(self, node: Node) -> None:
        if not isinstance(node, SignalNode):
            self.has_cpuif_reset_stack.pop()
            self.has_field_reset_stack.pop()



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .typing import PreElabRDLType

class ArrayedType():
    """
    Placeholder class to describe array types

    Once elaborated, arrays are converted to Python lists
    In the meantime, this placeholder is used to communicate expected type
    information during compilation type checking
    """
    def __init__(self, element_type: 'PreElabRDLType'):
        self.element_type = element_type

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ArrayedType):
            return self.element_type == other.element_type
        else:
            return NotImplemented



from typing import Type, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from .builtin_enums import BuiltinEnum
    from .user_enum import UserEnum
    from .references import PropertyReference, RefType
    from .user_struct import UserStruct
    from .array import ArrayedType
    from ..node import Node
    from .. import component as comp

# RDL Types the user will encounter via the public API
RDLValue = Union[
    int, bool, str, list,
    'BuiltinEnum', 'UserEnum', 'UserStruct', Type['UserEnum'],
    'Node', 'PropertyReference'
]

# RDL Types used internally prior to elaboration
PreElabRDLValue = Union[
    int, bool, str,
    'BuiltinEnum', 'UserEnum', 'UserStruct',
    'PropertyReference', 'comp.Component', 'RefType'
]

PreElabRDLType = Union[Type[PreElabRDLValue], 'ArrayedType']



from typing import TYPE_CHECKING, Any, Set, Type, Tuple

from .bases import PropertyRule

from .. import rdltypes
from .. import node as m_node

if TYPE_CHECKING:
    from .. import component as comp
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase
    from ..udp import UDPDefinition


class UserProperty(PropertyRule):
    """
    Base-class for user-defined properties
    """
    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def bindable_to(self) -> 'Set[Type[comp.Component]]': # type: ignore
        raise NotImplementedError

    @property
    def valid_type(self) -> Any:
        raise NotImplementedError

    @property
    def default_assignment(self) -> Any:
        raise NotImplementedError

    @property
    def constr_componentwidth(self) -> bool:
        raise NotImplementedError

    @property
    def valid_types(self) -> Tuple[Any, ...]: # type: ignore
        if isinstance(self.valid_type, rdltypes.ArrayedType):
            return (self.valid_type,)
        if issubclass(self.valid_type, rdltypes.references.RefType):
            return self.valid_type.expanded
        else:
            return (self.valid_type,)

    def get_name(self) -> str:
        return self.name

    def assign_value(self, comp_def: 'comp.Component', value: Any, src_ref: 'SourceRefBase') -> None:
        # Property assignments with no rhs show up as None here
        # For user-defined properties, this implies the default value
        # (15.2.2)
        if value is None:
            if self.default_assignment is None:
                # No default was set, so value is undefined.
                # As per 15.2.1-c. UDP is still "bound" to the component, so give
                # it the NoValue class as a value

                # Check if property is allowed in this component
                if type(comp_def) not in self.bindable_to:
                    self.env.msg.fatal(
                        "The property '%s' is not valid for '%s' components"
                        % (self.get_name(), type(comp_def).__name__.lower()),
                        src_ref
                    )

                # Circumvent the usual assignment function to avoid some pesky validation
                comp_def.properties[self.get_name()] = rdltypes.NoValue
                if src_ref is not None:
                    comp_def.property_src_ref[self.get_name()] = src_ref

                return

            value = self.default_assignment

        super().assign_value(comp_def, value, src_ref)

    def get_default(self, node: m_node.Node) -> Any:
        # If a user-defined property is not explicitly assigned, then it
        # does not get bound with its default value
        return None

    def validate(self, node: m_node.Node, value: Any) -> None:
        if self.constr_componentwidth:
            # 15.1.1-g: If constraint is set to componentwidth, the assigned
            #   value of the property shall not have a value of 1 for any bit
            #   beyond the width of the field.

            # Spec does not specify, but assuming this check gets ignored for
            # non-vector nodes
            if isinstance(node, m_node.VectorNode) and isinstance(value, int):
                if value.bit_length() > node.width:
                    self.env.msg.error(
                        "Value (%d) of the '%s' property cannot fit within the width (%d) of component '%s'"
                        % (value, self.name, node.width, node.inst_name),
                        node.inst.property_src_ref.get(self.name, node.inst.inst_src_ref)
                    )

        self._validate_ref_is_present(node, value)


class PureUserProperty(UserProperty):
    """
    UDP that was defined purely within SystemRDL source
    """

    def __init__(
        self, env: 'RDLEnvironment',
        name: str,
        valid_components: 'Set[Type[comp.Component]]',
        valid_type: Any,
        default_assignment: Any,
        constr_componentwidth: bool
    ):
        super().__init__(env)

        self._name = name
        self._valid_components = valid_components
        self._valid_type = valid_type
        self._default_assignment = default_assignment
        self._constr_componentwidth = constr_componentwidth

    @property
    def name(self) -> str:
        return self._name

    @property
    def bindable_to(self) -> 'Set[Type[comp.Component]]': # type: ignore
        return self._valid_components

    @property
    def valid_type(self) -> Any:
        return self._valid_type

    @property
    def default_assignment(self) -> Any:
        return self._default_assignment

    @property
    def constr_componentwidth(self) -> bool:
        return self._constr_componentwidth

class LegacyExternalUserProperty(PureUserProperty):
    pass

class ExternalUserProperty(UserProperty):
    """
    UDP that was defined via rdlc.register_udp().
    This variant fetches its definition contents from an external class
    """
    def __init__(self, env: 'RDLEnvironment', definition_cls: 'Type[UDPDefinition]', soft: bool=True):
        super().__init__(env)

        self.definition = definition_cls(env)
        self.is_soft = soft

    @property
    def name(self) -> str:
        return self.definition.name

    @property
    def bindable_to(self) -> 'Set[Type[comp.Component]]': # type: ignore
        return self.definition.valid_components

    @property
    def valid_type(self) -> Any:
        return self.definition.valid_type

    @property
    def default_assignment(self) -> Any:
        return self.definition.default_assignment

    @property
    def constr_componentwidth(self) -> bool:
        return self.definition.constr_componentwidth

    def get_default(self, node: m_node.Node) -> Any:
        return self.definition.get_unassigned_default(node)

    def validate(self, node: m_node.Node, value: Any) -> None:
        super().validate(node, value)
        self.definition.validate(node, value)



import hashlib
from typing import TYPE_CHECKING, List, Optional

from ..ast import ASTNode
from .helpers import is_pow2, roundup_pow2, roundup_to
from .value_normalization import normalize, RefInParameterError

from .. import component as comp
from .. import walker
from .. import rdltypes
from ..node import AddressableNode, RootNode, VectorNode, FieldNode, RegNode, RegfileNode
from ..node import AddrmapNode, MemNode, SignalNode, Node

if TYPE_CHECKING:
    from ..messages import MessageHandler
    from ..compiler import RDLEnvironment

#===============================================================================
# Elaboration Listeners
#===============================================================================

class ElabExpressionsListener(walker.RDLListener):
    """
    Elaborates all expressions
    - Component parameters
    - Instance array suffixes
    - Vector dimensions
    - Instance address allocators
    - Property assignments

    Also elaborates parameterized component type names
    """

    def __init__(self, msg_handler: 'MessageHandler'):
        self.msg = msg_handler


    def enter_Component(self, node: Node) -> None:
        # Evaluate component properties
        for prop_name, prop_value in node.inst.properties.items():
            if isinstance(prop_value, ASTNode):
                node.inst.properties[prop_name] = prop_value.get_value()


    def enter_AddressableComponent(self, node: AddressableNode) -> None:
        assert isinstance(node.inst, comp.AddressableComponent)
        # Evaluate instance object expressions
        if isinstance(node.inst.addr_offset, ASTNode):
            node.inst.addr_offset = node.inst.addr_offset.get_value()

        if isinstance(node.inst.addr_align, ASTNode):
            node.inst.addr_align = node.inst.addr_align.get_value()
            if node.inst.addr_align == 0:
                self.msg.fatal(
                    "Alignment allocator '%=' must be greater than zero",
                    node.inst.inst_src_ref
                )

        if node.inst.array_dimensions:
            for i, dim in enumerate(node.inst.array_dimensions):
                if isinstance(dim, ASTNode):
                    node.inst.array_dimensions[i] = dim.get_value()
                    if node.inst.array_dimensions[i] == 0:
                        self.msg.fatal(
                            "Array dimension must be greater than zero",
                            node.inst.inst_src_ref
                        )

        if isinstance(node.inst.array_stride, ASTNode):
            node.inst.array_stride = node.inst.array_stride.get_value()
            if node.inst.array_stride == 0:
                self.msg.fatal(
                    "Array stride allocator '+=' must be greater than zero",
                    node.inst.inst_src_ref
                )


    def enter_VectorComponent(self, node: VectorNode) -> None:
        assert isinstance(node.inst, comp.VectorComponent)
        # Evaluate instance object expressions
        if isinstance(node.inst.width, ASTNode):
            node.inst.width = node.inst.width.get_value()
            if node.inst.width == 0:
                self.msg.fatal(
                    "Vector width must be greater than zero",
                    node.inst.inst_src_ref
                )

        if isinstance(node.inst.msb, ASTNode):
            node.inst.msb = node.inst.msb.get_value()

        if isinstance(node.inst.lsb, ASTNode):
            node.inst.lsb = node.inst.lsb.get_value()




#-------------------------------------------------------------------------------
class PrePlacementValidateListener(walker.RDLListener):
    """
    Performs value checks of some properties prior to StructuralPlacementListener
    """
    def __init__(self, msg_handler: 'MessageHandler'):
        self.msg = msg_handler


    def enter_Addrmap(self, node: AddrmapNode) -> None:
        self.check_alignment(node)


    def enter_Regfile(self, node: RegfileNode) -> None:
        self.check_alignment(node)


    def check_alignment(self, node: AddressableNode) -> None:
        if 'alignment' in node.inst.properties:
            n = node.inst.properties['alignment']
            if n <= 0:
                self.msg.fatal(
                    "'alignment' property must be greater than zero",
                    node.inst.property_src_ref.get('alignment', node.inst.def_src_ref)
                )
            # 12.3.1-a, 13.4.1-b: All alignment values shall be a power of two (1, 2, 4, etc.)
            if not is_pow2(n):
                self.msg.fatal(
                    "'alignment' property must be a power of 2",
                    node.inst.property_src_ref.get('alignment', node.inst.def_src_ref)
                )


    def enter_Reg(self, node: RegNode) -> None:
        # 10.6.1-a: All registers shall have a regwidth = 2 N , where N >=3.
        if 'regwidth' in node.inst.properties:
            n = node.inst.properties['regwidth']
            if n < 8:
                self.msg.fatal(
                    "'regwidth' property must be at least 8",
                    node.inst.property_src_ref.get('regwidth', node.inst.def_src_ref)
                )
            if not is_pow2(n):
                self.msg.fatal(
                    "'regwidth' property must be a power of 2",
                    node.inst.property_src_ref.get('regwidth', node.inst.def_src_ref)
                )

        # 10.6.1-b: All registers shall have a accesswidth = 2 N , where N >=3.
        if 'accesswidth' in node.inst.properties:
            n = node.inst.properties['accesswidth']
            if n < 8:
                self.msg.fatal(
                    "'accesswidth' property must be at least 8",
                    node.inst.property_src_ref.get('accesswidth', node.inst.def_src_ref)
                )
            if not is_pow2(n):
                self.msg.fatal(
                    "'accesswidth' property must be a power of 2",
                    node.inst.property_src_ref.get('accesswidth', node.inst.def_src_ref)
                )


    def enter_Field(self, node: FieldNode) -> None:
        if 'fieldwidth' in node.inst.properties:
            n = node.inst.properties['fieldwidth']
            if n <= 0:
                self.msg.fatal(
                    "'fieldwidth' property must be greater than zero",
                    node.inst.property_src_ref.get('fieldwidth', node.inst.def_src_ref)
                )


    def enter_Signal(self, node: SignalNode) -> None:
        if 'signalwidth' in node.inst.properties:
            n = node.inst.properties['signalwidth']
            if n <= 0:
                self.msg.fatal(
                    "'signalwidth' property must be greater than zero",
                    node.inst.property_src_ref.get('signalwidth', node.inst.def_src_ref)
                )


    def enter_Mem(self, node: MemNode) -> None:
        # 11.3.1-a: mementries shall be greater than 0.
        if 'mementries' in node.inst.properties:
            n = node.inst.properties['mementries']
            if n <= 0:
                self.msg.fatal(
                    "'mementries' property must be greater than zero",
                    node.inst.property_src_ref.get('mementries', node.inst.def_src_ref)
                )

        # 11.3.1-a: memwidth shall be greater than 0.
        if 'memwidth' in node.inst.properties:
            n = node.inst.properties['memwidth']
            if n <= 0:
                self.msg.fatal(
                    "'memwidth' property must be greater than zero",
                    node.inst.property_src_ref.get('memwidth', node.inst.def_src_ref)
                )

#-------------------------------------------------------------------------------
class StructuralPlacementListener(walker.RDLListener):
    """
    Resolves inferred locations of structural components
    - Field width and offset
    - Component addresses
    - Signals.
    """

    def __init__(self, msg_handler: 'MessageHandler'):
        self.msg = msg_handler
        self.msb0_mode_stack = [] # type: List[bool]
        self.addressing_mode_stack = [] # type: List[rdltypes.AddressingType]
        self.alignment_stack = [] # type: List[Optional[int]]
        self.max_vreg_width = 0


    def enter_Addrmap(self, node: AddrmapNode) -> None:
        self.msb0_mode_stack.append(node.get_property('msb0'))
        self.addressing_mode_stack.append(node.get_property('addressing'))
        self.alignment_stack.append(node.get_property('alignment'))


    def enter_Regfile(self, node: RegfileNode) -> None:
        # Regfile can override the current alignment, but does not block
        # the propagation of a parent's setting if left undefined
        alignment = node.get_property('alignment')
        if alignment is None:
            # not set. Propagate from parent
            alignment = self.alignment_stack[-1]
        self.alignment_stack.append(alignment)


    def enter_Mem(self, node: MemNode) -> None:
        self.max_vreg_width = 0


    def exit_Mem(self, node: MemNode) -> None:
        # 11.3.1-d: memwidth defaults to regwidth
        # ... I assume that means if there are vregs inside a mem, then
        # memwidth inherits the max virtual reg regwidth?
        if 'memwidth' not in node.inst.properties:
            if self.max_vreg_width == 0:
                self.msg.error(
                    "Width of memory component is unknown. Either assign 'memwidth', or instantiate a virtual register to define the memory's width.",
                    node.inst.def_src_ref
                )
            node.inst.properties['memwidth'] = self.max_vreg_width

        self.resolve_addresses(node)


    def exit_Field(self, node: FieldNode) -> None:
        assert isinstance(node.inst, comp.Field)
        # Resolve field width
        if node.inst.width is None:
            fieldwidth = node.get_property('fieldwidth')

            if (node.inst.lsb is not None) and (node.inst.msb is not None):
                width = abs(node.inst.msb - node.inst.lsb) + 1

                node.inst.width = width
            elif fieldwidth is not None:
                node.inst.width = fieldwidth
            else:
                node.inst.width = 1

        # Test field width again
        fieldwidth = node.get_property('fieldwidth')
        if fieldwidth != node.inst.width:
            self.msg.fatal(
                "Width of field instance (%d) must match field's 'fieldwidth' property (%d)"
                % (node.inst.width, fieldwidth),
                node.inst.inst_src_ref
            )


    def exit_Signal(self, node: SignalNode) -> None:
        assert isinstance(node.inst, comp.Signal)

        # Resolve signal width
        if node.inst.width is None:
            signalwidth = node.get_property('signalwidth')

            if signalwidth is not None:
                node.inst.width = signalwidth
            else:
                node.inst.width = 1

        # Signals do not allow lsb/msb notation. Fill in values
        node.inst.lsb = 0
        node.inst.msb = node.inst.width - 1
        node.inst.low = 0
        node.inst.high = node.inst.width - 1

        # Test field width again
        signalwidth = node.get_property('signalwidth')
        if signalwidth != node.inst.width:
            self.msg.fatal(
                "Width of signal instance (%d) must match signal's 'signalwidth' property (%d)"
                % (node.inst.width, signalwidth),
                node.inst.inst_src_ref
            )


    def exit_Reg(self, node: RegNode) -> None:
        assert isinstance(node.inst, comp.Reg)

        regwidth = node.get_property('regwidth')

        self.max_vreg_width = max(regwidth, self.max_vreg_width)

        # Resolve field positions.
        # First determine if there is an implied lsb/msb mode
        implied_lsb_inst = None
        implied_msb_inst = None
        for inst in node.inst.children:
            if not isinstance(inst, comp.Field):
                continue

            if (inst.lsb is None) or (inst.msb is None):
                continue

            if inst.msb > inst.lsb:
                # bit ordering is [high:low]. Implies lsb mode
                implied_lsb_inst = inst
            elif inst.msb < inst.lsb:
                # bit ordering is [low:high]. Implies msb mode
                implied_msb_inst = inst

        # 10.7.1-a: Both the [low:high] and [high:low] bit specification forms
        #   shall not be used together in the same register.
        if (implied_lsb_inst is not None) and (implied_msb_inst is not None):
            # register uses both [high:low] and [low:high] ordering!
            self.msg.fatal(
                "Both the [low:high] (field '%s') and [high:low] (field '%s') bit specification forms shall not be used together in the same register."
                % (implied_msb_inst.inst_name, implied_lsb_inst.inst_name),
                node.inst.def_src_ref
            )

        # Any implied lsb/msb modes override the property set by a parent
        if implied_msb_inst is not None:
            node.inst.is_msb0_order = True
        elif implied_lsb_inst is not None:
            node.inst.is_msb0_order = False
        else:
            node.inst.is_msb0_order = self.msb0_mode_stack[-1]

        # Assign field positions
        # Children are iterated in order of declaration
        prev_inst = None # type: Optional[comp.Field]
        for inst in node.inst.children:
            if not isinstance(inst, comp.Field):
                continue

            if (inst.lsb is None) or (inst.msb is None):
                # Offset is not known

                if node.env.chk_implicit_field_pos:
                    node.env.msg.message(
                        node.env.chk_implicit_field_pos,
                        "Bit offset for field '%s' is not explicit" % inst.inst_name,
                        inst.inst_src_ref
                    )

                if node.inst.is_msb0_order:
                    # In msb0 mode. Pack from top first
                    # lsb == high
                    # msb == low
                    if prev_inst is None:
                        inst.lsb = regwidth - 1
                    else:
                        inst.lsb = prev_inst.msb - 1

                    inst.msb = inst.lsb - inst.width + 1

                    if inst.msb < 0:
                        node.env.msg.fatal(
                            "Field '%s' of width %d infers bit range [%d:%d] which extends below bit 0"
                            % (inst.inst_name, inst.width, inst.msb, inst.lsb),
                            inst.inst_src_ref
                        )
                else:
                    # In lsb0 mode. Pack from bit 0 first
                    # lsb == low
                    # msb == high
                    if prev_inst is None:
                        inst.lsb = 0
                    else:
                        inst.lsb = prev_inst.msb + 1

                    inst.msb = inst.lsb + inst.width - 1
            inst.high = max(inst.msb, inst.lsb)
            inst.low = min(inst.msb, inst.lsb)
            prev_inst = inst

        # Sort fields by low-bit.
        # Non-field child components are sorted to be first (signals)
        def get_field_sort_key(inst: comp.Component) -> int:
            if not isinstance(inst, comp.Field):
                return -1
            else:
                return inst.low
        node.inst.children.sort(key=get_field_sort_key)


    def exit_Regfile(self, node: RegfileNode) -> None:
        self.resolve_addresses(node)

        self.alignment_stack.pop()


    def exit_Addrmap(self, node: AddrmapNode) -> None:
        is_bridge = node.get_property('bridge')
        self.resolve_addresses(node, is_bridge)

        self.msb0_mode_stack.pop()
        self.addressing_mode_stack.pop()
        self.alignment_stack.pop()


    def exit_AddressableComponent(self, node: AddressableNode) -> None:
        assert isinstance(node.inst, comp.AddressableComponent)
        # Resolve array stride if needed
        if node.inst.is_array and (node.inst.array_stride is None):
            node.inst.array_stride = node.size

            if node.env.chk_implicit_addr:
                node.env.msg.message(
                    node.env.chk_implicit_addr,
                    "Array stride of component '%s' is not explicitly set" % node.inst.inst_name,
                    node.inst.inst_src_ref
                )


    def resolve_addresses(self, node: AddressableNode, is_bridge: bool = False) -> None:
        """
        Resolve addresses of children of Addrmap and Regfile components
        """

        # Get alignment based on 'alignment' property
        # This remains constant for all children
        prop_alignment = self.alignment_stack[-1]
        if prop_alignment is None:
            # was not specified. Does not contribute to alignment
            prop_alignment = 1

        prev_node = None
        for child_node in node.children(skip_not_present=False):
            if not isinstance(child_node, AddressableNode):
                continue
            assert isinstance(child_node.inst, comp.AddressableComponent)

            if child_node.inst.addr_offset is not None:
                # Address is already known. Do not need to infer
                # Still, check that it honors the requested alignment
                if (child_node.raw_address_offset % prop_alignment) != 0:
                    self.msg.error(
                        "Address offset +0x%x of component '%s' must be aligned on a 0x%x byte boundary"
                        % (child_node.raw_address_offset, child_node.inst_name, prop_alignment),
                        child_node.inst.inst_src_ref
                    )
                prev_node = child_node
                continue

            if node.env.chk_implicit_addr:
                node.env.msg.message(
                    node.env.chk_implicit_addr,
                    "Address offset of component '%s' is not explicitly set" % child_node.inst.inst_name,
                    child_node.inst.inst_src_ref
                )

            # Get alignment specified by '%=' allocator, if any
            alloc_alignment = child_node.inst.addr_align
            if alloc_alignment is None:
                # was not specified. Does not contribute to alignment
                alloc_alignment = 1

            # Calculate alignment based on current addressing mode
            if self.addressing_mode_stack[-1] == rdltypes.AddressingType.compact:
                if isinstance(child_node, RegNode):
                    # Regs are aligned based on their accesswidth
                    mode_alignment = child_node.get_property('accesswidth') // 8
                else:
                    # Spec does not specify for other components
                    # Assuming absolutely compact packing
                    mode_alignment = 1

            elif self.addressing_mode_stack[-1] == rdltypes.AddressingType.regalign:
                # Components are aligned to a multiple of their size
                # Spec vaguely suggests that alignment is also a power of 2
                mode_alignment = child_node.size
                mode_alignment = roundup_pow2(mode_alignment)

            elif self.addressing_mode_stack[-1] == rdltypes.AddressingType.fullalign:
                # Same as regalign except for arrays
                # Arrays are aligned to their total size
                # Both are rounded to power of 2
                mode_alignment = child_node.total_size
                mode_alignment = roundup_pow2(mode_alignment)

            else:
                raise RuntimeError

            # Calculate resulting address offset
            alignment = max(prop_alignment, alloc_alignment, mode_alignment)
            if (prev_node is None) or is_bridge:
                next_offset = 0
            else:
                next_offset = prev_node.raw_address_offset + prev_node.total_size

            # round next_offset up to alignment
            child_node.inst.addr_offset = roundup_to(next_offset, alignment)

            prev_node = child_node

        # Sort children by address offset
        # Non-addressable child components are sorted to be first (signals)
        def get_child_sort_key(inst: comp.Component) -> int:
            if not isinstance(inst, comp.AddressableComponent):
                return -1
            else:
                return inst.addr_offset
        node.inst.children.sort(key=get_child_sort_key)

#-------------------------------------------------------------------------------
class LateElabListener(walker.RDLListener):
    """
    Elaboration listener for misc late-stage things
    """
    def __init__(self, msg_handler: 'MessageHandler', env: 'RDLEnvironment'):
        self.msg = msg_handler
        self.env = env
        self.coerce_external_to = None # type: Optional[bool]
        self.coerce_end_regfile = None # type: Optional[Node]

        self.node_needs_revisit = [] # type: List[Node]


    def enter_Component(self, node: Node) -> None:
        # if parent is not present, children inherit it too
        if node.parent is not None and not isinstance(node.parent, RootNode):
            if not node.parent.get_property('ispresent'):
                node.inst.properties['ispresent'] = False


    def enter_Field(self, node: FieldNode) -> None:
        # Inherits internal/external of parent reg
        assert node.parent is not None
        node.inst.external = node.parent.inst.external


    def enter_Regfile(self, node: RegfileNode) -> None:
        if self.coerce_external_to is not None:
            # Is nested inside another regfile that is coercing to a particular inst type
            node.inst.external = self.coerce_external_to
        elif node.inst.external is not None:
            # First regfile to specify inst type. Coerce all children to the same type
            # as per 12.2-f
            self.coerce_external_to = node.inst.external
            self.coerce_end_regfile = node
        else:
            # Regfile did not specify internal/external
            # Set to default of internal
            node.inst.external = False


    def enter_Reg(self, node: RegNode) -> None:
        assert isinstance(node.inst, comp.Reg)

        if self.coerce_external_to is not None:
            # Is nested inside regfile that is coercing to a particular inst type
            node.inst.external = self.coerce_external_to
        elif node.inst.external is None:
            if node.inst.is_alias:
                # inherit internal/external instance type from alias primary
                assert isinstance(node.inst.alias_primary_inst, comp.Reg)
                if node.inst.alias_primary_inst.external is not None:
                    node.inst.external = node.inst.alias_primary_inst.external
                else:
                    # Elaborate did not resolve the primary reg yet and it is still unspecified.
                    # re-visit this reg later
                    self.node_needs_revisit.append(node)
            else:
                node.inst.external = False

        # Register aliases with their primary register
        if node.inst.is_alias:
            assert isinstance(node.inst.alias_primary_inst, comp.Reg)
            node.inst.alias_primary_inst._alias_names.append(node.inst_name)


    def exit_Regfile(self, node: RegfileNode) -> None:
        if node is self.coerce_end_regfile:
            # Exiting inst type coercion
            self.coerce_external_to = None
            self.coerce_end_regfile = None


    def exit_Component(self, node: Node) -> None:
        # Generate elaborated type name
        # (only if it exists. Some importers will not assign a type name)
        if node.inst.type_name is not None:
            extra_type_name_segments = []

            # Augment based on parameter overrides as per 5.1.1.4
            if node.inst.original_def is not None:
                for i, inst_parameter in enumerate(node.inst.parameters):
                    orig_param_value = node.inst.original_def.parameters[i].get_value()
                    new_param_value = inst_parameter.get_value()
                    if new_param_value != orig_param_value:
                        try:
                            segment = inst_parameter.get_normalized_parameter()
                            extra_type_name_segments.append(segment)
                        except RefInParameterError:
                            self.msg.error(
                                "Parameter '%s' contains a reference. SystemRDL does not allow component references inside parameter values."
                                % inst_parameter.name,
                                node.inst.inst_src_ref
                            )

            # Further augment type name as per extended type generation from DPAs
            if self.env.use_extended_type_name_gen:
                # Strip duplicates and sort alphabetically
                node.inst._dyn_assigned_props = sorted(set(node.inst._dyn_assigned_props))
                node.inst._dyn_assigned_children = sorted(set(node.inst._dyn_assigned_children))

                # assignments made 'through' the component
                for child_name in node.inst._dyn_assigned_children:
                    child = node.inst.get_child_by_name(child_name)
                    assert child is not None

                    # <child_name>_<hash of child type name>
                    if child.type_name is not None:
                        norm_name = hashlib.new('md5', child.type_name.encode('utf-8'), usedforsecurity=False).hexdigest()[:8]
                    else:
                        # an external importer did not assign a type name.
                        # Use the inst name instead
                        norm_name = hashlib.new('md5', child.inst_name.encode('utf-8'), usedforsecurity=False).hexdigest()[:8]
                    extra_type_name_segments.append(child_name + "_" + norm_name)

                # this component's DPAs
                for prop_name in node.inst._dyn_assigned_props:
                    # <prop_name>_<norm prop value>
                    norm_name = normalize(node.get_property(prop_name), owner_node=node)
                    extra_type_name_segments.append(prop_name + "_" + norm_name)

            if extra_type_name_segments:
                node.inst.type_name += "_" + "_".join(extra_type_name_segments)


#-------------------------------------------------------------------------------
class LateElabRevisitor:
    """
    In rare situations, some nodes need to be re-visited one last time in order
    to complete elaboration.

    Rather than re-traversing the entire register model, these are set aside and are
    re-visited here.
    """

    def __init__(self, revisit_nodes: List[Node]) -> None:
        for node in revisit_nodes:
            if isinstance(node, RegNode):
                self.revisit_reg(node)
            else:
                raise RuntimeError

    def revisit_reg(self, node: RegNode) -> None:

        # Resolve alias register's instance type if it was not possible to do so earlier.
        # By now, its primary would have been resolved.
        if node.inst.external is None:
            assert isinstance(node.inst, comp.Reg)
            if node.inst.is_alias:
                assert isinstance(node.inst.alias_primary_inst, comp.Reg)
                if node.inst.alias_primary_inst.external is not None:
                    node.inst.external = node.inst.alias_primary_inst.external
                else:
                    raise RuntimeError



from typing import Optional
from enum import IntEnum

from .node import AddressableNode, VectorNode, FieldNode, RegNode, RegfileNode
from .node import AddrmapNode, MemNode, SignalNode
from .node import RootNode, Node


class WalkerAction(IntEnum):

    #: Continue walking the register model
    Continue = 0

    #: Walker will continue calling listener methods for this component, but
    #: will not recurse into this node's children.
    SkipDescendants = 1

    #: Stop the walker immediately. No more listener methods will be called.
    StopNow = 2


class RDLListener:
    """
    Base class for user-defined RDL traversal listeners.

    From each callback, optionally return a :class:`WalkerAction` to
    control how the walker should continue model traversal.
    Returning ``None`` is equivalent to :attr:`WalkerAction.Continue`.


    .. versionchanged:: 1.23
        Added optional WalkerAction return value
    """
    def enter_Component(self, node: Node) -> Optional[WalkerAction]:
        pass

    def exit_Component(self, node: Node) -> Optional[WalkerAction]:
        pass

    def enter_AddressableComponent(self, node: AddressableNode) -> Optional[WalkerAction]:
        pass

    def exit_AddressableComponent(self, node: AddressableNode) -> Optional[WalkerAction]:
        pass

    def enter_VectorComponent(self, node: VectorNode) -> Optional[WalkerAction]:
        pass

    def exit_VectorComponent(self, node: VectorNode) -> Optional[WalkerAction]:
        pass

    def enter_Addrmap(self, node: AddrmapNode) -> Optional[WalkerAction]:
        pass

    def exit_Addrmap(self, node: AddrmapNode) -> Optional[WalkerAction]:
        pass

    def enter_Regfile(self, node: RegfileNode) -> Optional[WalkerAction]:
        pass

    def exit_Regfile(self, node: RegfileNode) -> Optional[WalkerAction]:
        pass

    def enter_Mem(self, node: MemNode) -> Optional[WalkerAction]:
        pass

    def exit_Mem(self, node: MemNode) -> Optional[WalkerAction]:
        pass

    def enter_Reg(self, node: RegNode) -> Optional[WalkerAction]:
        pass

    def exit_Reg(self, node: RegNode) -> Optional[WalkerAction]:
        pass

    def enter_Field(self, node: FieldNode) -> Optional[WalkerAction]:
        pass

    def exit_Field(self, node: FieldNode) -> Optional[WalkerAction]:
        pass

    def enter_Signal(self, node: SignalNode) -> Optional[WalkerAction]:
        pass

    def exit_Signal(self, node: SignalNode) -> Optional[WalkerAction]:
        pass

#===============================================================================
class RDLWalker:
    """
    Implements a walker instance that traverses the elaborated RDL instance tree
    Each node is visited exactly once.

    Each node is visited as follows:

    1. Run :func:`~RDLListener.enter_Component` callback
    2. Run :func:`~RDLListener.enter_AddressableComponent` or :func:`~RDLListener.enter_VectorComponent` callback
    3. Run type-specific ``enter_*()`` callback, such as :func:`~RDLListener.enter_Reg`
    4. Traverse any children
    5. Run type-specific ``exit_*()`` callback, such as :func:`~RDLListener.exit_Reg`
    6. Run :func:`~RDLListener.exit_AddressableComponent` or :func:`~RDLListener.exit_VectorComponent` callback
    7. Run :func:`~RDLListener.exit_Component` callback

    """
    def __init__(self, unroll: bool=False, skip_not_present: bool=True):
        """
        Parameters
        ----------
        unroll : bool
            If True, any nodes that are arrays are unrolled.
            When the walker arrives at an array node, it will be visited multiple
            times according to the array dimensions.

        skip_not_present : bool
            If True, walker skips nodes whose 'ispresent' property is set
            to False
        """
        self.unroll = unroll
        self.skip_not_present = skip_not_present
        self.current_action = WalkerAction.Continue


    def walk(self, node: Node, *listeners: RDLListener, skip_top: bool=False) -> None:
        """
        Initiates the walker to traverse the current ``node`` and its children.
        Calls the corresponding callback for each of the ``listeners`` provided in
        the order that they are listed.

        Parameters
        ----------
        node : :class:`~systemrdl.node.Node`
            Node to start traversing.
            Listener traversal includes this node.

        listeners : :class:`~RDLListener`
            One or more :class:`~RDLListener` that are invoked during
            node traversal.
            Listener callbacks are executed in the same order as provided.

        skip_top : bool
            Skip callbacks for the top node specified by ``node``


        .. versionchanged:: 1.21
            Added ``skip_top`` option.
        """

        if not skip_top:
            for listener in listeners:
                self.current_action = self.do_enter(node, listener)
                if self.current_action == WalkerAction.StopNow:
                    return

        if self.current_action == WalkerAction.SkipDescendants:
            # skip recursion into children, then reset action
            self.current_action = WalkerAction.Continue
        else:
            for child in node.children(unroll=self.unroll, skip_not_present=self.skip_not_present):
                self.walk(child, *listeners)
                if self.current_action == WalkerAction.StopNow:
                    return

        if not skip_top:
            for listener in listeners:
                self.current_action = self.do_exit(node, listener)
                if self.current_action == WalkerAction.StopNow:
                    return


    def do_enter(self, node: Node, listener: RDLListener) -> WalkerAction:
        action = WalkerAction.Continue
        new_action = WalkerAction.Continue

        # Skip RootNode since it isn't really a component
        if not isinstance(node, RootNode):
            action = listener.enter_Component(node) or WalkerAction.Continue

        if action == WalkerAction.StopNow:
            return action

        if isinstance(node, AddressableNode):
            new_action = listener.enter_AddressableComponent(node) or WalkerAction.Continue
        elif isinstance(node, VectorNode):
            new_action = listener.enter_VectorComponent(node) or WalkerAction.Continue

        action = max(new_action, action)
        if action == WalkerAction.StopNow:
            return action

        if isinstance(node, FieldNode):
            new_action = listener.enter_Field(node) or WalkerAction.Continue
        elif isinstance(node, RegNode):
            new_action = listener.enter_Reg(node) or WalkerAction.Continue
        elif isinstance(node, RegfileNode):
            new_action = listener.enter_Regfile(node) or WalkerAction.Continue
        elif isinstance(node, AddrmapNode):
            new_action = listener.enter_Addrmap(node) or WalkerAction.Continue
        elif isinstance(node, MemNode):
            new_action = listener.enter_Mem(node) or WalkerAction.Continue
        elif isinstance(node, SignalNode):
            new_action = listener.enter_Signal(node) or WalkerAction.Continue

        action = max(new_action, action)

        return action


    def do_exit(self, node: Node, listener: RDLListener) -> WalkerAction:
        action = WalkerAction.Continue
        new_action = WalkerAction.Continue

        if isinstance(node, FieldNode):
            action = listener.exit_Field(node) or WalkerAction.Continue
        elif isinstance(node, RegNode):
            action = listener.exit_Reg(node) or WalkerAction.Continue
        elif isinstance(node, RegfileNode):
            action = listener.exit_Regfile(node) or WalkerAction.Continue
        elif isinstance(node, AddrmapNode):
            action = listener.exit_Addrmap(node) or WalkerAction.Continue
        elif isinstance(node, MemNode):
            action = listener.exit_Mem(node) or WalkerAction.Continue
        elif isinstance(node, SignalNode):
            action = listener.exit_Signal(node) or WalkerAction.Continue

        if action == WalkerAction.StopNow:
            return action

        if isinstance(node, AddressableNode):
            new_action = listener.exit_AddressableComponent(node) or WalkerAction.Continue
        elif isinstance(node, VectorNode):
            new_action = listener.exit_VectorComponent(node) or WalkerAction.Continue

        action = max(new_action, action)
        if action == WalkerAction.StopNow:
            return action

        # Skip RootNode since it isn't really a component
        if not isinstance(node, RootNode):
            new_action = listener.exit_Component(node) or WalkerAction.Continue

        action = max(new_action, action)
        return action



from typing import Optional, Dict, Any, TYPE_CHECKING, List, Type, Generator
from collections import OrderedDict
import inspect
import copyreg


from ..core import rdlformatcode
from .. import component as comp

if TYPE_CHECKING:
    from markdown import Markdown



class UserEnumMemberContainer:
    """
    Container class used only when defining a new UserEnum.
    This shall only be used by importers.


    .. versionadded:: 1.24
    """
    def __init__(self, name: str, value: int, rdl_name: Optional[str] = None, rdl_desc: Optional[str] = None) -> None:
        self.name = name
        self.value = value
        self.rdl_name = rdl_name
        self.rdl_desc = rdl_desc

    def __hash__(self) -> int:
        return hash((self.name, self.value, self.rdl_name, self.rdl_desc))

class UserEnumMeta(type):
    """
    Metaclass for UserEnum
    """

    _member_map = {} # type: Dict[str, UserEnum]
    _parent_scope = None # type: Optional[comp.Component]

    def __bool__(cls) -> bool:
        # classes/types should always be True.
        return True

    def __contains__(cls, obj: Any) -> bool:
        if not isinstance(obj, UserEnum):
            raise TypeError(
                f"unsupported operand type(s) for 'in': '{type(obj).__qualname__}' and '{cls.__class__.__qualname__}'"
            )
        return isinstance(obj, cls) and obj._name in cls._member_map

    def __getitem__(cls, name: str) -> 'UserEnum':
        return cls._member_map[name]

    def __iter__(cls) -> 'Generator[UserEnum, None, None]':
        return (member for member in cls._member_map.values())

    def __len__(cls) -> int:
        return len(cls._member_map)

    def __repr__(cls) -> str:
        return f"<UserEnum {cls.__name__}>"

    def __reversed__(cls) -> 'Generator[UserEnum, None, None]':
        """
        Returns members in reverse definition order.
        """
        return (member for member in reversed(cls._member_map.values()))

    @property
    def __members__(cls) -> Dict[str, 'UserEnum']:
        return cls._member_map

    @property
    def members(cls) -> Dict[str, 'UserEnum']:
        """
        Returns a mapping of member name->value.
        """
        return cls._member_map

    @property
    def type_name(cls) -> str:
        # (docstring is in rst)
        return cls.__name__


    def define_new(cls, name: str, members: List[UserEnumMemberContainer], _parent_scope: Optional[comp.Component]=None) -> Type['UserEnum']:
        """
        Define a new UserEnum class

        Parameters
        ----------
        name: str
            Name of the enum type
        members: UserEnumMemberContainer
            List of enum members


        .. versionadded:: 1.24
        """

        # Validate all member names and values are unique
        names = [m.name for m in members]
        values = [m.value for m in members]
        if len(names) != len(set(names)):
            raise ValueError("All members of an enum shall have unique names")
        if len(values) != len(set(values)):
            raise ValueError("All members of an enum shall have unique values")

        # Create the new class
        classdict = {
            '_member_map': OrderedDict(),
            '_parent_scope': _parent_scope,
        }
        metaclass = cls.__class__
        newcls = metaclass(name, (cls,), classdict)

        # populate enum members, which are each instances of the new class
        for m in members:
            enum_member = newcls(
                m.name, m.value, m.rdl_name, m.rdl_desc
            )
            newcls._member_map[m.name] = enum_member

        return newcls # type: ignore

    def get_parent_scope(cls) -> Optional[comp.Component]:
        """
        Returns reference to parent component that contains this type definition.
        """
        return cls._parent_scope

    def get_scope_path(cls, scope_separator: str="::") -> str:
        """
        Generate a string that represents this enum's declaration namespace
        scope.

        Parameters
        ----------
        scope_separator: str
            Override the separator between namespace scopes
        """
        parent_scope = cls.get_parent_scope() # pylint: disable=no-value-for-parameter
        if parent_scope is None:
            # Importer likely never set the scope
            return ""
        elif isinstance(parent_scope, comp.Root):
            # Declaration was in root scope
            return ""
        else:
            # Get parent definition's scope path
            parent_path = parent_scope.get_scope_path(scope_separator)

            # Extend it with its scope name
            if parent_path:
                return(
                    parent_path
                    + scope_separator
                    + parent_scope._scope_name
                )
            else:
                return parent_scope._scope_name

class UserEnum(metaclass=UserEnumMeta):
    """
    All user-defined enum types are extended from this class and can be
    identified using ``issubclass(obj, UserEnum)``. Enum members are instances
    of this class and can be identified using ``isinstance(obj, UserEnum)``.
    """

    def __init__(self, name: str, value: int, rdl_name: Optional[str], rdl_desc: Optional[str]) -> None:
        self._name = name
        self._value = value
        self._rdl_name = rdl_name
        self._rdl_desc = rdl_desc

    def __repr__(self) -> str:
        return "<%s.%s: %r>" % (
            self.__class__.__name__, self._name, self._value
        )

    def __str__(self) -> str:
        return "%s.%s" % (
            self.__class__.__name__, self._name
        )

    def __int__(self) -> int:
        return self.value

    def __bool__(self) -> bool:
        return bool(self.value)

    def __deepcopy__(self, memo: Dict[int, Any]) -> 'UserEnum':
        # Do not deepcopy enumerations
        return self

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> int:
        return self._value

    @property
    def rdl_name(self) -> Optional[str]:
        """
        Enum entry's ``name`` property
        """
        return self._rdl_name

    @property
    def rdl_desc(self) -> Optional[str]:
        """
        Enum entry's ``desc`` property
        """
        return self._rdl_desc

    def get_html_desc(self, markdown_inst: Optional['Markdown']=None) -> Optional[str]:
        """
        Translates the enum's 'desc' property into HTML.

        Any RDLFormatCode tags used are converted to HTML.
        The text is also fed through a Markdown processor.

        The additional Markdown processing allows designers the choice to use a
        more modern lightweight markup language as an alternative to SystemRDL's
        "RDLFormatCode".

        Parameters
        ----------
        markdown_inst: ``markdown.Markdown``
            Override the class instance of the Markdown processor.
            See the `Markdown module <https://python-markdown.github.io/reference/#Markdown>`_
            for more details.

        Returns
        -------
        str or None
            HTML formatted string.
            If enum entry does not have a description, returns ``None``


        .. versionchanged:: 1.6
            Added ``markdown_inst`` option.
        """
        desc_str = self._rdl_desc
        if desc_str is None:
            return None
        return rdlformatcode.rdlfc_to_html(desc_str, md=markdown_inst)

    def get_html_name(self) -> Optional[str]:
        """
        Translates the enum's 'name' property into HTML.

        Any RDLFormatCode tags used are converted to HTML.

        Returns
        -------
        str or None
            HTML formatted string.
            If enum entry does not have an explicitly set name, returns ``None``


        .. versionadded:: 1.8
        """
        name_str = self._rdl_name
        if name_str is None:
            return None
        return rdlformatcode.rdlfc_to_html(name_str, is_desc=False)


# Tell pickle how to reduce dynamically generated UserEnum classes
def _reduce_user_enum(c: Type[UserEnum]) -> Any:
    if c is UserEnum:
        # Reached base class. Return string so pickle can look up the actual object
        return 'UserEnum'

    assert len(c.__bases__) == 1 # Only supporting single-inheritence
    base_cls = c.__bases__[0] # type: Type[UserEnum]

    # decompose members back into factory containers
    members = []
    for member in c.members.values():
        m = UserEnumMemberContainer(
            member._name, member._value, member._rdl_name, member._rdl_desc
        )
        members.append(m)

    args = (c.type_name, members, c._parent_scope)
    return (base_cls.define_new, args)
copyreg.pickle(UserEnumMeta, _reduce_user_enum) # type: ignore


def is_user_enum(t: Any) -> bool:
    """
    Test if type ``t`` is a :class:`~UserEnum`

    .. note:: Returns false if ``t`` is referencing a UserEnum value member
    """
    return inspect.isclass(t) and issubclass(t, UserEnum)



from typing import TYPE_CHECKING, Optional, Any

from .ast_node import ASTNode

from .cast import is_castable

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase
    from rdltypes.typing import PreElabRDLType

    OptionalSourceRef = Optional[SourceRefBase]

# conditional operator
#   i ? j : k
# Truth expression is self-determined and does not contribute to context

class Conditional(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', i: ASTNode, j: ASTNode, k: ASTNode):
        super().__init__(env, src_ref)
        self.i = i
        self.j = j
        self.k = k
        self.is_numeric = None # type: bool

    def predict_type(self) -> 'PreElabRDLType':
        t_i = self.i.predict_type()
        if not is_castable(t_i, bool):
            self.msg.fatal(
                "Conditional operand of expression is not a compatible boolean type",
                self.src_ref
            )

        # Type of j and k shall be compatible
        t_j = self.j.predict_type()
        t_k = self.k.predict_type()

        typ = None # type: Any
        if is_castable(t_j, int) and is_castable(t_k, int):
            self.is_numeric = True
            typ = int
        elif t_j == t_k:
            # Same types. Inherently compatible
            self.is_numeric = False
            typ = t_j
        else:
            # Incompatible
            self.msg.fatal(
                "True/False results of ternary conditional are not compatible types",
                self.src_ref
            )
        return typ

    def get_min_eval_width(self) -> int:
        # Truth operand has no influence in evaluation context
        return(max(
            self.j.get_min_eval_width(),
            self.k.get_min_eval_width()
        ))

    def get_value(self, eval_width: Optional[int]=None) -> Any:
        # i is self-determined
        i = bool(self.i.get_value())

        if self.is_numeric:
            if eval_width is None:
                eval_width = max(
                    self.j.get_min_eval_width(),
                    self.k.get_min_eval_width()
                )
            j = self.j.get_value(eval_width)
            k = self.k.get_value(eval_width)
        elif not self.is_numeric:
            j = self.j.get_value()
            k = self.k.get_value()
        else:
            raise RuntimeError

        if i:
            return j
        else:
            return k



import hashlib
from typing import Any, Optional, Union, List

from .. import rdltypes
from .. import node

class RefInParameterError(Exception):
    pass

def normalize(value: Any, owner_node: Optional[node.Node]=None) -> str:
    """
    Flatten an RDL value into a unique string that is used for type
    normalization.
    """
    # Determine what type is being flattened
    if isinstance(value, bool):
        return normalize_boolean(value)
    elif isinstance(value, int):
        return normalize_scalar(value)
    elif isinstance(value, str):
        return normalize_string(value)
    elif isinstance(value, list):
        return normalize_array(value, owner_node)
    elif isinstance(value, (rdltypes.BuiltinEnum, rdltypes.UserEnum)):
        return normalize_enum(value)
    elif isinstance(value, rdltypes.UserStruct):
        return normalize_struct(value, owner_node)
    elif isinstance(value, node.Node):
        return normalize_component_ref(value, owner_node)
    elif isinstance(value, rdltypes.PropertyReference):
        return normalize_property_ref(value, owner_node)
    elif rdltypes.is_user_enum(value):
        return normalize_user_enum_type(value)
    elif value is rdltypes.NoValue:
        # this only happens if a UDP is assigned with no value
        return "NaN"
    elif isinstance(value, rdltypes.references.ComponentRef) and owner_node is None:
        # This only happens if a parameter contains a struct that wraps a reference
        # datatype. # Since the parameter's value doesnt bother getting resolved,
        # It shows up as a ComponentRef object here.
        # SystemRDL spec does not allow references in parameters, so this ends up
        # being an odd, but convenient place to catch this.
        raise RefInParameterError
    else:
        # Should never get here
        raise RuntimeError(value, owner_node)


def normalize_scalar(value: int) -> str:
    """
    5.1.1.4 - c.1:
        Scalar values shall be rendered using their hexadecimal representation.
    """
    return f"{value:x}"


def normalize_boolean(value: bool) -> str:
    """
    5.1.1.4 - c.2:
        Boolean values shall be rendered using either t for true or f for false.
    """
    if value:
        return "t"
    else:
        return "f"


def normalize_string(value: str) -> str:
    """
    5.1.1.4 - c.3:
        String values shall be rendered using the first eight characters of
        their md5 (Message-Digest Algorithm) checksum.
    """
    md5 = hashlib.new('md5', value.encode('utf-8'), usedforsecurity=False).hexdigest()
    return md5[:8]


def normalize_enum(value: Union[rdltypes.BuiltinEnum, rdltypes.UserEnum]) -> str:
    """
    5.1.1.4 - c.4:
        Enum values shall be rendered using their enumerator literal.
    """
    return value.name


def normalize_array(value: List[Any], owner_node: Optional[node.Node]=None) -> str:
    """
    5.1.1.4 - c.5:
        Arrays shall be rendered by:
        1. generating the normalized values of its elements,
        2. joining these elements with single underscores (_) into a single
            character sequence, and
        3. using the first eight characters of the md5 checksum of this
            character sequence

        ... which can be semi-formalized as:
            subsequence( md5( join( normalized_values, '_' ), 0, 8 )
    """
    norm_elements = []
    for element in value:
        norm_elements.append(normalize(element, owner_node))

    norm_str = "_".join(norm_elements)
    md5 = hashlib.new('md5', norm_str.encode('utf-8'), usedforsecurity=False).hexdigest()
    return md5[:8]


def normalize_struct(value: rdltypes.UserStruct, owner_node: Optional[node.Node]=None) -> str:
    """
    5.1.1.4 - c.6:
        Structs shall be rendered by:
        1. generating the normalized value of each member,
        2. joining each member’s name with its normalized value, separated by
            a single underscore (_),
        3. joining the member character sequences with single underscores,
        4. using the first eight characters of the md5 checksum of this
            character sequence

        ... which can be semi-formalized as:
            member_normalization = concat( member_name, '_', normalized_member_value )
            subsequence( md5( join( apply( struct_members, member_normalization ) ), 0, 8)
    """
    norm_elements = []
    for member_name, member_value in value._values.items():
        norm_elements.append(f"{member_name}_{normalize(member_value, owner_node)}")

    norm_str = "_".join(norm_elements)
    md5 = hashlib.new('md5', norm_str.encode('utf-8'), usedforsecurity=False).hexdigest()
    return md5[:8]


def normalize_component_ref(value: node.Node, owner_node: node.Node) -> str:
    """
    Hash of relative path from owner of the property to the target component
    """
    path = value.get_rel_path(owner_node)
    md5 = hashlib.new('md5', path.encode('utf-8'), usedforsecurity=False).hexdigest()
    return md5[:8]


def normalize_property_ref(value: rdltypes.PropertyReference, owner_node: node.Node) -> str:
    """
    Hash of relative path from owner of the property to the target component's
    property
    """
    path = f"{value.node.get_rel_path(owner_node)}->{value.name}"
    md5 = hashlib.new('md5', path.encode('utf-8'), usedforsecurity=False).hexdigest()
    return md5[:8]


def normalize_user_enum_type(value: type) -> str:
    """
    Enum type references shall be rendered using their enumeration type name.
    """
    return value.__name__



# Explicitly import these to ensure their contents are loaded, otherwise subclass
# discovery wont find their contents since nothing else actually imports them directly.
from . import builtin
from . import prop_refs



from typing import List, Type, TYPE_CHECKING
import inspect

from ..parser.SystemRDLParser import SystemRDLParser

from .BaseVisitor import BaseVisitor
from .ExprVisitor import ExprVisitor
from .helpers import get_ID_text

from ..properties.user_defined import PureUserProperty
from .. import ast

from ..source_ref import src_ref_from_antlr
from .. import component as comp
from .. import rdltypes

if TYPE_CHECKING:
    from typing import Optional, Union, Any, Set
    from ..compiler import RDLCompiler

class UDPVisitor(BaseVisitor):
    def __init__(self, compiler: 'RDLCompiler') -> None:
        super().__init__(compiler)

        self._constr_componentwidth = None # type: Optional[bool]
        self._default_assignment = None # type: Any
        self._valid_type = None # type: Optional[Type[Union[int, str, bool, rdltypes.BuiltinEnum, rdltypes.UserEnum, rdltypes.UserStruct, comp.Component, rdltypes.references.RefType]]]
        self._bindable_to = None # type: Optional[Set[Type[comp.Component]]]

    def visitUdp_def(self, ctx: SystemRDLParser.Udp_defContext) -> None:
        udp_name = get_ID_text(ctx.ID())

        # Collect all attributes
        for attr_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Udp_attrContext):
            self.visit(attr_ctx)

        # Check attribute rules
        # 15.1.1.b: A user-defined property definition shall include the
        #   component property specification.
        if self._valid_type is None:
            self.msg.fatal(
                "User-defined property '%s' does not specify the 'type' attribute"
                % udp_name,
                src_ref_from_antlr(ctx.ID())
            )

        # 15.1.1.c: A user-defined property definition shall include its type definition
        if self._bindable_to is None:
            self.msg.fatal(
                "User-defined property '%s' does not specify the 'component' attribute"
                % udp_name,
                src_ref_from_antlr(ctx.ID())
            )

        # 15.1 Table 30: Currently limited to componentwidth for type bit
        if self._constr_componentwidth:
            if self._valid_type != int:
                self.msg.fatal(
                    "Constraint 'componentwidth' is only valid for properties of type 'bit'",
                    src_ref_from_antlr(ctx.ID())
                )

        # Evaluate default value, if any
        if self._default_assignment is not None:
            expr_ctx = self._default_assignment

            visitor = ExprVisitor(self.compiler)
            expr = visitor.visit(expr_ctx)
            expr_type = expr.predict_type()

            # temporarily expand valid_type into an array
            if inspect.isclass(self._valid_type) and issubclass(self._valid_type, rdltypes.references.RefType):
                valid_types = self._valid_type.expanded
            else:
                valid_types = (self._valid_type,) # type: ignore

            # 15.1.1-e: The default value shall be assignment compatible with
            # the property type
            for valid_type in valid_types:
                if ast.is_castable(expr_type, valid_type):
                    if isinstance(expr, ast.ASTNode):
                        # Found a type-compatible match. (first match is best match)
                        # Wrap the expression with an explicit assignment cast
                        expr = ast.AssignmentCast(
                            self.compiler.env, src_ref_from_antlr(expr_ctx),
                            expr, valid_type
                        )
                    break
            else:
                self.msg.fatal(
                    "Property default is incompatible with property type",
                    src_ref_from_antlr(expr_ctx)
                )

            self._default_assignment = expr

        constr_componentwidth = self._constr_componentwidth
        if constr_componentwidth is None:
            constr_componentwidth = False

        # create UDP
        udp = PureUserProperty(
            self.compiler.env,
            udp_name,
            self._bindable_to,
            self._valid_type,
            self._default_assignment,
            constr_componentwidth
        )
        self.compiler.env.property_rules.register_udp(udp, src_ref_from_antlr(ctx.ID()))


    def visitUdp_type(self, ctx: SystemRDLParser.Udp_typeContext) -> None:
        # Determine which type this property can hold
        if self._valid_type is not None:
            self.msg.fatal(
                "More than one 'type' attribute specified for user-defined property",
                src_ref_from_antlr(ctx.TYPE_kw())
            )

        token = self.visit(ctx.udp_data_type())
        valid_type = self.datatype_from_token(token)

        is_array = ctx.array_type_suffix() is not None
        if is_array:
            # arrayify
            valid_type = rdltypes.ArrayedType(valid_type) # type: ignore

        self._valid_type = valid_type


    _UDPUsage_Map = {
        SystemRDLParser.FIELD_kw    : comp.Field,
        SystemRDLParser.REG_kw      : comp.Reg,
        SystemRDLParser.REGFILE_kw  : comp.Regfile,
        SystemRDLParser.ADDRMAP_kw  : comp.Addrmap,
        SystemRDLParser.MEM_kw      : comp.Mem,
        SystemRDLParser.SIGNAL_kw   : comp.Signal,
        #SystemRDLParser.CONSTRAINT_kw   : TODO,
    }
    def visitUdp_usage(self, ctx: SystemRDLParser.Udp_usageContext) -> None:
        # Determine which components the UDP can be bound to
        if self._bindable_to is not None:
            self.msg.fatal(
                "More than one 'component' attribute specified for user-defined property",
                src_ref_from_antlr(ctx.COMPONENT_kw())
            )

        comp_types = [] # type: List[Type[comp.Component]]
        for token_ctx in ctx.getTypedRuleContexts(SystemRDLParser.Udp_comp_typeContext):
            token = self.visit(token_ctx)
            if token.type == SystemRDLParser.ALL_kw:
                comp_types.extend(list(self._UDPUsage_Map.values()))
            else:
                comp_types.append(self._UDPUsage_Map[token.type])

        self._bindable_to = set(comp_types)


    def visitUdp_default(self, ctx: SystemRDLParser.Udp_defaultContext) -> None:
        if self._default_assignment is not None:
            self.msg.fatal(
                "More than one 'default' attribute specified for user-defined property",
                src_ref_from_antlr(ctx.DEFAULT_kw())
            )

        # defer expr evaluation until later
        self._default_assignment = ctx.expr()


    def visitUdp_constraint(self, ctx: SystemRDLParser.Udp_constraintContext) -> None:
        # There is only one type of constraint even allowed by the grammar
        # no need to explore further
        if self._constr_componentwidth is not None:
            self.msg.fatal(
                "More than one 'constraint' attribute specified for user-defined property",
                src_ref_from_antlr(ctx.CONSTRAINT_kw())
            )

        self._constr_componentwidth = True



from .ast_node import ASTNode

from .literals import BoolLiteral, IntLiteral, StringLiteral
from .literals import BuiltinEnumLiteral, EnumLiteral
from .literals import StructLiteral, ArrayLiteral, ExternalLiteral

from .references import ParameterRef, ArrayIndex, MemberRef, InstRef, PropRef

from .relational import Eq, Neq, Lt, Gt, Leq, Geq

from .sequence import Concatenate, Replicate

from .binary import Add, Sub, Mult, Div, Mod
from .binary import BitwiseAnd, BitwiseOr, BitwiseXor, BitwiseXnor

from .unary import UnaryPlus, UnaryMinus, BitwiseInvert

from .reduction import AndReduce, NandReduce, OrReduce, NorReduce, XorReduce, XnorReduce, BoolNot

from .exponential import Exponent, LShift, RShift

from .boolean import BoolAnd, BoolOr

from .conditional import Conditional

from .cast import WidthCast, BoolCast, AssignmentCast, is_castable



from typing import TYPE_CHECKING, Optional, Type

from .ast_node import ASTNode
from .conditional import is_castable

from ..core.helpers import truncate_int

if TYPE_CHECKING:
    from ..compiler import RDLEnvironment
    from ..source_ref import SourceRefBase

    OptionalSourceRef = Optional[SourceRefBase]

# Integer binary operators:
#   +  -  *  /  %  &  |  ^  ^~  ~^
# Normal expression context rules
class _BinaryIntExpr(ASTNode):
    def __init__(self, env: 'RDLEnvironment', src_ref: 'OptionalSourceRef', l: ASTNode, r: ASTNode):
        super().__init__(env, src_ref)
        self.l = l
        self.r = r

    def predict_type(self) -> Type[int]:
        l_type = self.l.predict_type()
        r_type = self.r.predict_type()
        if not is_castable(l_type, int):
            self.msg.fatal(
                "Left operand of expression is not a compatible numeric type",
                self.src_ref
            )
        if not is_castable(r_type, int):
            self.msg.fatal(
                "Right operand of expression is not a compatible numeric type",
                self.src_ref
            )
        return int

    def get_min_eval_width(self) -> int:
        return(max(
            self.l.get_min_eval_width(),
            self.r.get_min_eval_width()
        ))

class Add(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l + r, eval_width)

class Sub(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l - r, eval_width)

class Mult(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l * r, eval_width)

class Div(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))

        if r == 0:
            self.msg.fatal(
                "Division by zero",
                self.src_ref
            )
        return truncate_int(l // r, eval_width)

class Mod(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))

        if r == 0:
            self.msg.fatal(
                "Modulo by zero",
                self.src_ref
            )
        return truncate_int(l % r, eval_width)

class BitwiseAnd(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l & r, eval_width)

class BitwiseOr(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l | r, eval_width)

class BitwiseXor(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l ^ r, eval_width)

class BitwiseXnor(_BinaryIntExpr):
    def get_value(self, eval_width: Optional[int]=None) -> int:
        if eval_width is None:
            eval_width = self.get_min_eval_width()
        l = int(self.l.get_value(eval_width))
        r = int(self.r.get_value(eval_width))
        return truncate_int(l ^~ r, eval_width)
